"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations
from uuid import uuid4
import logging

from pydantic import field_validator, Field, ValidationInfo
from pydantic.dataclasses import dataclass

from .circuit import Circuit, Channel
from .utilities.validation_tools import (
    pauli_error,
    dataclass_config,
    retrieve_field,
    no_name_error,
)

logging.basicConfig(format="%(name)s - %(levelname)s - %(message)s")
log = logging.getLogger(__name__)


def default_circuit(pauli_str: str) -> Circuit:
    """
    Construct the default circuit corresponding to the input Pauli string.

    Parameters
    ----------
    pauli_str: str
        The input Pauli string of the stabilizer

    Returns
    -------
    Circuit
        The circuit containing instructions measuring the stabilizer
    """
    weight = len(pauli_str)
    ancilla_channel = Channel(type="quantum", label="a0")
    data_channels = [Channel(type="quantum", label=f"d{i}") for i in range(weight)]
    cbit_channel = Channel(type="classical", label="c0")

    reset = [Circuit("Reset_0", channels=[ancilla_channel])]
    hadamard1 = [Circuit("H", channels=[ancilla_channel])]
    hadamard2 = [Circuit("H", channels=[ancilla_channel])]
    entangle_ancilla = [
        [Circuit(f"C{pauli}", channels=[ancilla_channel, data_channels[i]])]
        for i, pauli in enumerate(pauli_str)
    ]
    measurement = [Circuit("Measurement", channels=[ancilla_channel, cbit_channel])]
    ops_list = [reset, hadamard1] + entangle_ancilla + [hadamard2, measurement]

    circuit_name = f"stab_{pauli_str}"
    return Circuit(
        circuit_name,
        channels=data_channels + [ancilla_channel, cbit_channel],
        circuit=ops_list,
    )


@dataclass(config=dataclass_config)
class SyndromeCircuit:
    """
    A SyndromeCircuit object specifies one way to measure a pauli string.
    There can be multiple SyndromeCircuit objects for the same pauli string,
    specifying different circuits for how to measure it.
    The qubits in the `circuit` field are labeled 0, ..., n.
    Once the actual syndrome extraction circuit for a stabilizer is generated,
    qubit i of the SyndromeCircuit object is replaced by stabilizer.data_qubits[i].
    Therefore the qubit order inside stabilizer.data_qubits must match the
    order in which they are used in SyndromeCircuit.

    Parameter
    ---------
    pauli : str
        The pauli string which this circuit is supposed to measure. Must be a string
    name: str, optional
        A name for the SyndromeCircuit object consisting of the characters {X, Y, Z}.
    circuit: Circuit | None
        The circuit instructions to measure the syndrome. If `None` is provided, a
        default circuit for the given pauli string is automatically constructed.
    uuid: str
        A uuid associated with the object
    """

    pauli: str = Field(min_length=1)
    name: str = Field(min_length=1, default_factory=lambda: "SyndromeCircuit")
    circuit: Circuit | None = Field(default_factory=lambda: None, validate_default=True)
    uuid: str = Field(default_factory=lambda: str(uuid4()), validate_default=True)

    # Validation functions
    _validate_pauli = field_validator("pauli")(pauli_error)
    _validate_name = field_validator("name")(no_name_error)

    @field_validator("circuit", mode="after")
    @classmethod
    def default_syndrome_circuit(cls, value, info: ValidationInfo):
        """
        In case the field circuit is not provided, the default
        syndrome extraction circuit is automatically assigned
        """
        pauli_str = retrieve_field("pauli", info)
        value = default_circuit(pauli_str) if value is None else value
        return value

    # Magic methods
    def __eq__(self, other: SyndromeCircuit) -> bool:
        """
        Ignore the uuid in the equality check.
        """
        if self.pauli != other.pauli:
            log.info("The two circuits measure different Pauli strings.")
            log.debug("%s != %s\n", self.pauli, other.pauli)
            return False

        if self.name != other.name:
            log.info("The two circuits have different names.")
            log.debug("%s != %s\n", self.name, other.name)
            return False

        if self.circuit != other.circuit:
            log.info("The two circuits have different circuit instructions.")
            return False
            # log.debug(
            #     f"\n{self.circuit}\n  !=\n{other.circuit}"
            # )

        return True

    def __hash__(self):
        """
        Ignore the uuid in hashing.
        """
        return hash((self.pauli, self.name, self.circuit))



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=too-many-lines

from __future__ import annotations
from itertools import combinations, product, chain
from functools import cached_property
from uuid import uuid4

import numpy as np

from pydantic.dataclasses import Field, dataclass
from pydantic import field_validator, model_validator, ValidationInfo
from pydantic_core import ArgsKwargs

from .stabilizer import Stabilizer
from .pauli_operator import PauliOperator
from .syndrome_circuit import SyndromeCircuit
from .matrices import ParityCheckMatrix
from .tanner_graphs import TannerGraph
from .utilities.pauli_format_conversion import paulixz_to_char
from .utilities.stab_array import (
    merge_stabarrays,
    reduce_stabarray_with_bookkeeping,
    find_destabarray,
    StabArray,
    invert_bookkeeping_matrix,
)
from .utilities.validation_tools import (
    dataclass_config,
    ensure_tuple,
    empty_list_error,
    retrieve_field,
)


@dataclass(config=dataclass_config)
class Block:  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """
    Block describing one or multiple logical qubits. It is defined by a code type, a
    list of stabilizers, two logical operators per logical qubit and a unique label. The
    block also contains information about how its syndromes are measured.

    Parameters
    ----------
    stabilizers : tuple[Stabilizer, ...]
        The stabilizers that define the block.
    logical_x_operators : tuple[PauliOperator, ...]
        The logical X operators associated to the block.
    logical_z_operators : tuple[PauliOperator, ...]
        The logical Z operators associated to the block.
    syndrome_circuits : tuple[SyndromeCircuit, ...]
        The syndrome circuits with which the stabilizers of this block are measured.
        NOTE: Block does not have a check for distinct Syndrome Circuits(s)
    stabilizer_to_circuit : dict[str, str]
        A dictionary mapping stabilizer uuids to the uuids of the syndrome circuits that
        measure them.
    unique_label : str, optional
        Label for the block. It must be unique among all blocks in the initial Eka.
        If no label is provided, a unique label is generated automatically using the
        uuid module.
    skip_validation : bool, optional
        Boolean that allows to skip some validation of the Block. Default is False.
        The validation skipped are the following:

        - The qubits coordinates have the same dimension.
        - The stabilizers commute with each other.
        - The logical operators commute with each other.
        - The stabilizers commute with the logical operators.
        - The logical X and Z operators anti-commute at the same index and commute at \
        different indices.
        - The number of qubits and stabilizers in the Block is compatible with the \
        number of logical qubits.
    uuid : str, optional
        Unique identifier for the block. If no uuid is provided, a unique uuid is
        generated automatically using the uuid module.

    Attributes
    ----------
    __version__ : str
        Version of the Block class implementation.
    """

    stabilizers: tuple[Stabilizer, ...]
    logical_x_operators: tuple[PauliOperator, ...]
    logical_z_operators: tuple[PauliOperator, ...]
    syndrome_circuits: tuple[SyndromeCircuit, ...] = Field(
        default_factory=tuple, validate_default=True
    )
    stabilizer_to_circuit: dict[str, str] = Field(
        default_factory=dict, validate_default=True
    )
    unique_label: str = Field(default_factory=lambda: str(uuid4()))
    skip_validation: bool = Field(default=False, validate_default=True)
    uuid: str = Field(default_factory=lambda: str(uuid4()), validate_default=True)

    # version of the Block class implementation
    __version__: str = "1.0.0"

    # Model validators with mode="before".
    # Note that these are executed in the reverse order in which they are defined.
    # This is where the last "before" validator is executed
    @model_validator(mode="before")
    @classmethod
    def _validate_qubits_included(cls, data: dict):
        """
        Check that all qubits used in the logical operators are included in the
        set of stabilizers.

        Parameters
        ----------
        data : dict
            The data to be validated.
        """
        logical_operator_qubits = set(
            qubit
            for operator in data["logical_x_operators"] + data["logical_z_operators"]
            for qubit in operator.data_qubits
        )
        stabilizers_qubits = set(
            qubit
            for stabilizer in data["stabilizers"]
            for qubit in stabilizer.data_qubits
        )
        qubits_not_in_stabilizers = logical_operator_qubits - stabilizers_qubits
        if qubits_not_in_stabilizers:
            raise ValueError(
                f"Qubits {qubits_not_in_stabilizers} are not included in the"
                f" stabilizers but are used in the logical operators"
            )
        return data

    @model_validator(mode="before")
    @classmethod
    def _validate_number_of_logical_operators(cls, data: dict):
        """
        Check that the number of logical X operators is equal to the number of logical Z
        """
        if len(data["logical_x_operators"]) != len(data["logical_z_operators"]):
            raise ValueError(
                "The number of logical X operators must be equal to the number of "
                "logical Z operators."
            )
        return data

    @model_validator(mode="before")
    @classmethod
    def _assign_types(cls, data: ArgsKwargs):
        """
        Assign the types of the stabilizers and logical operators and casts
        ArgsKwargs into a dictionary. This is necessary to ensure that the data is
        loaded in the correct format when the Block is created from a JSON. It also
        tests for empty inputs for stabilizers, logical_x_operators and
        logical_z_operators.
        """
        # Cast ArgsKwargs into a dictionary
        if not isinstance(data, dict):
            data = {**dict(data.kwargs), **dict(data.args)}
        # Check empty arguments
        list(
            map(
                empty_list_error,
                [
                    data["stabilizers"],
                    data["logical_x_operators"],
                    data["logical_z_operators"],
                ],
            )
        )
        # Assign types explicitly
        if isinstance(data["stabilizers"][0], dict):
            data["stabilizers"] = [Stabilizer(**stab) for stab in data["stabilizers"]]
        if isinstance(data["logical_x_operators"][0], dict):
            data["logical_x_operators"] = [
                PauliOperator(**{k: v for k, v in op.items() if k != "nr_of_ancillae"})
                for op in data["logical_x_operators"]
            ]
        if isinstance(data["logical_z_operators"][0], dict):
            data["logical_z_operators"] = [
                PauliOperator(**{k: v for k, v in op.items() if k != "nr_of_ancillae"})
                for op in data["logical_z_operators"]
            ]
        return data

    @model_validator(mode="before")
    @classmethod
    def _validate_workbench_json_version(cls, data):
        """
        Checks that the major version of the incoming Workbench JSON matches
        the major version of the Block class.
        """
        if hasattr(data, "kwargs") and data.kwargs is not None:
            if "__version__" not in data.kwargs:
                return data
            if data.kwargs["__version__"][0] != cls.__version__[0]:
                raise ValueError(
                    """
                    The major version of the Workbench export is not 
                    compatible with the major version of Eka.Block.
                    """
                )
        return data

    # Field validators are executed after model_validator with mode="before"
    @field_validator("stabilizers", mode="before")
    @classmethod
    def _validate_distinct_stabilizers(cls, stabilizers: tuple[Stabilizer, ...]):
        """
        Check that stabilizers are distinct.
        """
        for stab1, stab2 in combinations(stabilizers, 2):
            if stab1 == stab2:
                raise ValueError("Stabilizers must be distinct.")
        return stabilizers

    _validate_stabilizers_list = field_validator("stabilizers", mode="before")(
        ensure_tuple
    )

    @field_validator("logical_x_operators", mode="before")
    @classmethod
    def _validate_distinct_logical_x_operators(
        cls, logical_x_operators: tuple[PauliOperator, ...]
    ):
        """
        Check that logical X operators are distinct.
        """
        for log1, log2 in combinations(logical_x_operators, 2):
            if log1 == log2:
                raise ValueError("Logical X operators must be distinct.")
        return logical_x_operators

    _validate_logical_x_list = field_validator("logical_x_operators", mode="before")(
        ensure_tuple
    )

    @field_validator("logical_z_operators", mode="before")
    @classmethod
    def _validate_distinct_logical_z_operators(
        cls, logical_z_operators: tuple[PauliOperator, ...]
    ):
        """
        Check that logical Z operators are distinct.
        """
        for log1, log2 in combinations(logical_z_operators, 2):
            if log1 == log2:
                raise ValueError("Logical Z operators must be distinct.")
        return logical_z_operators

    _validate_logical_z_list = field_validator("logical_z_operators", mode="before")(
        ensure_tuple
    )

    @field_validator("syndrome_circuits", mode="after")
    @classmethod
    def _create_default_syndrome_circuits(
        cls, syndrome_circuits: tuple[SyndromeCircuit, ...], info: ValidationInfo
    ) -> dict[str, str]:
        """
        Check that for every pauli string of the stabilizers there is a corresponding
        syndrome circuit. If not, create a default syndrome circuit for that pauli
        string.
        """
        stab_paulis = {stab.pauli for stab in retrieve_field("stabilizers", info)}
        synd_circ_paulis = {synd_circ.pauli for synd_circ in syndrome_circuits}
        additional_circuits = [
            SyndromeCircuit(
                name=f"default_{pauli}",
                pauli=pauli,
            )
            for pauli in stab_paulis - synd_circ_paulis
        ]
        return tuple(list(syndrome_circuits) + additional_circuits)

    @field_validator("stabilizer_to_circuit", mode="after")
    @classmethod
    def _check_stabilizer_to_circuit_map_uuids(
        cls, stabilizer_to_circuit: dict[str, str], info: ValidationInfo
    ) -> dict[str, str]:
        """
        Check that all uuids in the stabilizer to circuit map are valid.
        I.e. check that all stabilizer uuids actually appear in the set of stabilizers
        and that all syndrome circuit uuids actually appear in the list of syndrome
        circuits. Also check that pauli strings match.
        """
        syndrome_circs_dict = {
            circ.uuid: circ for circ in retrieve_field("syndrome_circuits", info)
        }
        stabilizer_dict = {
            stab.uuid: stab for stab in retrieve_field("stabilizers", info)
        }
        for stab_uuid, syndrome_circ_uuid in stabilizer_to_circuit.items():
            if stab_uuid not in stabilizer_dict.keys():
                raise ValueError(
                    f"Stabilizer with uuid {stab_uuid} is not present in the "
                    "stabilizers."
                )
            if syndrome_circ_uuid not in syndrome_circs_dict.keys():
                raise ValueError(
                    f"Syndrome circuit with uuid {syndrome_circ_uuid} is not present "
                    f"in the syndrome circuits."
                )
            if (
                stabilizer_dict[stab_uuid].pauli
                != syndrome_circs_dict[syndrome_circ_uuid].pauli
            ):
                raise ValueError(
                    f"Stabilizer with uuid {stab_uuid} has a different pauli string "
                    f"{stabilizer_dict[stab_uuid].pauli} than the syndrome circuit "
                    f"with uuid {syndrome_circ_uuid} (pauli string: "
                    f"{syndrome_circs_dict[syndrome_circ_uuid].pauli})."
                )
        return stabilizer_to_circuit

    @field_validator("stabilizer_to_circuit", mode="after")
    @classmethod
    def _associate_stabs_to_syndrome_circuits(
        cls, stabilizer_to_circuit: dict[str, str], info: ValidationInfo
    ) -> dict[str, str]:
        """
        For every stabilizer which is not yet associated to a syndrome circuit,
        associate it with a default syndrome circuit. If there are multiple syndrome
        circuits for the same pauli string, raise an exception.
        """
        synd_circs = retrieve_field("syndrome_circuits", info)
        for stab in retrieve_field("stabilizers", info):
            if stab.uuid not in stabilizer_to_circuit.keys():
                matching_circuits = [
                    circ for circ in synd_circs if circ.pauli == stab.pauli
                ]
                if len(matching_circuits) > 1:
                    raise ValueError(
                        "Multiple syndrome circuits for the same stabilizer pauli "
                        f"string {stab.pauli} found. Could not automatically associate "
                        f"stabilizer {stab} with a syndrome circuit. Please do the "
                        "association manually."
                    )
                stabilizer_to_circuit[stab.uuid] = matching_circuits[0].uuid
        return stabilizer_to_circuit

    # Model validators, mode="after"
    # NOTE that the order of the validators is important
    @model_validator(mode="after")
    def _validate_coordinate_dimension(self):
        """
        Check that all qubits coordinates have the same dimension.
        """
        # Bypass validation
        if self.skip_validation:
            return self

        dimension = len(self.qubits[0])
        if any(len(coordinate) != dimension for coordinate in self.qubits):
            raise ValueError("All qubits coordinates must have the same dimension.")
        return self

    @model_validator(mode="after")
    def _validate_commutation_stabilizers(self):
        """
        Check that stabilizers commute with each other.
        """
        # Bypass validation
        if self.skip_validation:
            return self

        for stab1, stab2 in combinations(self.stabilizers, 2):
            if not stab1.commutes_with(stab2):
                raise ValueError(
                    f"Stabilizers must commute with each other:\n"
                    f"{stab1} and\n"
                    f"{stab2} do not commute.\n"
                )
        return self

    @model_validator(mode="after")
    def _validate_commutation_logical_operators(self):
        """
        Check that logical operators commute with each other.
        """
        # Bypass validation
        if self.skip_validation:
            return self

        for log1, log2 in combinations(self.logical_x_operators, 2):
            if not log1.commutes_with(log2):
                raise ValueError(
                    f"Logical X operators must commute with each other:\n"
                    f"{log1} and\n"
                    f"{log2} do not commute.\n"
                )
        for log1, log2 in combinations(self.logical_z_operators, 2):
            if not log1.commutes_with(log2):
                raise ValueError(
                    f"Logical Z operators must commute with each other:\n"
                    f"{log1} and\n"
                    f"{log2} do not commute.\n"
                )
        return self

    @model_validator(mode="after")
    def _validate_commutation_stabilizers_logical_operators(self):
        """
        Check that stabilizers commute with logical operators.
        """
        # Bypass validation
        if self.skip_validation:
            return self

        for stab, log_x in product(self.stabilizers, self.logical_x_operators):
            if not stab.commutes_with(log_x):
                raise ValueError(
                    f"Stabilizers must commute with logical X operators:\n"
                    f"{repr(stab)} and\n"
                    f"{repr(log_x)} do not commute.\n"
                )
        for stab, log_z in product(self.stabilizers, self.logical_z_operators):
            if not stab.commutes_with(log_z):
                raise ValueError(
                    f"Stabilizers must commute with logical Z operators:\n"
                    f"{repr(stab)} and\n"
                    f"{repr(log_z)} do not commute.\n"
                )
        return self

    @model_validator(mode="after")
    def _validate_anticommutation_logical_operators_one_to_one(self):
        """
        Check that the logical X and Z operators anti-commute at the same index
        anticommute and commute at different indices.
        """
        # Bypass validation
        if self.skip_validation:
            return self

        for (i, log_x), (j, log_z) in combinations(
            chain(
                enumerate(self.logical_x_operators), enumerate(self.logical_z_operators)
            ),
            2,
        ):
            if i == j and log_x.commutes_with(log_z):
                raise ValueError(
                    f"Logical X and Z operators at the same index must anticommute "
                    f"with each other:\n{log_x} and\n{log_z} do not anticommute.\n"
                )
            if i != j and not log_x.commutes_with(log_z):
                raise ValueError(
                    f"Logical X and Z operators at different indices must commute with "
                    f"each other:\n{log_x} and\n {log_z} do not commute.\n"
                )
        return self

    # This is the last "after" validator
    @model_validator(mode="after")
    def _validate_dimensional_compatibility(self):
        """
        Check that the number of qubits and stabilizers in the Block is compatible
        with the number of logical qubits.

        E.g. if L is the number of logical operators, N the number of data qubits and k
        the number of independent stabilizers, then L = N - k.
        """
        # Bypass validation
        if self.skip_validation:
            return self

        if self.reduced_stabarray.nstabs + len(self.logical_x_operators) != len(
            self.data_qubits
        ):
            raise ValueError(
                "The number of qubits and independent stabilizers in the Block is "
                "not compatible with the number of logical qubits."
            )
        return self

    # Properties
    @property
    def data_qubits(self) -> tuple[tuple[int, ...], ...]:
        """
        Return a tuple of all data qubits in the block.

        Returns
        -------
        tuple[tuple[int, ...], ...] :
            A tuple of coordinates representing the data qubits.
        """
        return tuple(
            sorted(
                # sort the qubits so that the order is not dependent on the
                # order of the stabilizers
                set(
                    data_qubit
                    for stabilizer in self.stabilizers
                    for data_qubit in stabilizer.data_qubits
                )
            )
        )

    @property
    def ancilla_qubits(self) -> tuple[tuple[int, ...], ...]:
        """
        Return the set of all ancilla qubits in the block.

        Returns
        -------
        tuple[tuple[int, ...], ...] :
            A tuple of coordinates representing the ancilla qubits.
        """
        return tuple(
            set(
                ancilla_qubit
                for stabilizer in self.stabilizers
                for ancilla_qubit in stabilizer.ancilla_qubits
            )
        )

    @property
    def qubits(self) -> tuple[tuple[int, ...], ...]:
        """
        Return the set of all qubits in the block.

        Returns
        -------
        tuple[tuple[int, ...], ...] :
            A tuple of coordinates representing all qubits in the block.
        """
        return tuple(set(self.data_qubits + self.ancilla_qubits))

    @property
    def n_data_qubits(self) -> int:
        """
        Return the number of data qubits in the block.
        """
        return len(self.data_qubits)

    @property
    def n_logical_qubits(self) -> int:
        """
        Return the number of logical qubits in the block.
        """
        return len(self.logical_x_operators)

    @property
    def n_irreducible_stabs(self) -> int:
        """
        Return the number of irreducible stabilizers in the embedding, i.e. the
        minimum number of stabilizers needed to generate the code space.
        """
        return self.reduced_stabarray.nstabs

    @cached_property
    def original_stabarray(self) -> StabArray:
        """
        Return the stabilizer array of the block as defined by the stabilizers.
        """
        signed_pauli_ops = [
            s.as_signed_pauli_op(self.data_qubits) for s in self.stabilizers
        ]
        return StabArray.from_signed_pauli_ops(signed_pauli_ops, validated=False)

    @cached_property
    def reduced_stabarray_with_bookkeeping(self) -> tuple[StabArray, np.ndarray]:
        """
        Return the reduced stabilizer array of the embedding with bookkeeping.
        """
        return reduce_stabarray_with_bookkeeping(self.original_stabarray)

    @property
    def reduced_stabarray(self) -> StabArray:
        """
        Return the reduced stabilizer array of the embedding.
        """
        return self.reduced_stabarray_with_bookkeeping[0]

    @property
    def bookkeeping(self) -> np.ndarray:
        """
        Return the bookkeeping array of the reduced stabilizer array.
        """
        return self.reduced_stabarray_with_bookkeeping[1]

    @property
    def reduced_bookkeeping(self) -> np.ndarray:
        """
        Return the reduced bookkeeping array of the reduced stabilizer array. This
        entails slicing out the last rows of the bookkeeping array that should cancel
        out and give a zero row.
        """
        return self.bookkeeping[: self.n_irreducible_stabs, :]

    @cached_property
    def bookkeeping_inv(self) -> np.ndarray:
        """
        Return the inverted bookkeeping array of the reduced stabilizer array.
        """
        return invert_bookkeeping_matrix(self.reduced_stabarray_with_bookkeeping[1])

    @property
    def reduced_bookkeeping_inv(self) -> np.ndarray:
        """
        Return the inverted reduced bookkeeping array of the reduced stabilizer
        array. This entails slicing out the last trivial columns of the inverted
        bookkeeping array that correspond to the zero rows of the reduced StabArray.
        """
        return self.bookkeeping_inv[:, : self.n_irreducible_stabs]

    @cached_property
    def x_log_stabarray(self) -> StabArray:
        """
        Return the X stabilizer array of the logical operator set.
        """
        signed_pauli_ops = [
            x.as_signed_pauli_op(self.data_qubits) for x in self.logical_x_operators
        ]
        return StabArray.from_signed_pauli_ops(signed_pauli_ops, validated=False)

    @cached_property
    def z_log_stabarray(self) -> StabArray:
        """
        Return the Z stabilizer array of the logical operator set.
        """
        signed_pauli_ops = [
            z.as_signed_pauli_op(self.data_qubits) for z in self.logical_z_operators
        ]
        return StabArray.from_signed_pauli_ops(signed_pauli_ops, validated=False)

    @cached_property
    def destabarray(self) -> StabArray:
        """
        Return the destabilizer array of the block. The operators in the
        destabarray anti-commute with exactly one stabilizer of the block each. In
        particular, the one with a matching index. They also commute with all of the
        logical operators. Note that the destabilizer is calculated using the reduced
        stabilizer array and not the original one.
        """
        z_state_stabarray = merge_stabarrays(
            (self.reduced_stabarray, self.z_log_stabarray)
        )
        full_destabarray = find_destabarray(
            z_state_stabarray, partial_destabarray=self.x_log_stabarray
        )
        # The last n_logical_qubits rows are the destabilizers of the logical operators
        # and as such they should be omitted from the full_destabarray.
        return StabArray.from_signed_pauli_ops(
            full_destabarray[: -self.n_logical_qubits]
        )

    @cached_property
    def pauli_charges(self) -> dict[tuple[int, ...], str]:
        """
        Calculate Pauli charges for all data qubits in the given Block. The Pauli
        charges are calculated from the stabilizers of the Block. For every data qubit,
        one counts how often the data qubit is included in stabilizers in the X, Y, and
        Z basis respectively. If it is included in an odd number of X, Y, or Z
        stabilizers, the data qubit has a Pauli charge of X, Y, or Z respectively. We
        only report a single Pauli charge where multiple charges are combined into one
        charge according to the product of Pauli matrices. E.g. if a data qubit has a
        Pauli charge of both X and Z, the combined Pauli charge is Y since Y=iXZ.

        E.g. for a d=5 rotated surface code, plotting the Pauli charges on top of the
        data qubits would look like this (where we omitted plotting data qubits with no
        Pauli charge):

        .. code-block::

            Y -- X -- X -- X -- Y
            |    |    |    |    |
            Z --   --   --   -- Z
            |    |    |    |    |
            Z --   --   --   -- Z
            |    |    |    |    |
            Z --   --   --   -- Z
            |    |    |    |    |
            Y -- X -- X -- X -- Y

        In this example of the rotated surface code, the four logical corners have Pauli
        charge Y, and there are two boundaries with Pauli charge X and two with Pauli
        charge Z. The bulk data qubits have no Pauli charge. No Pauli charge is
        represented by "_" in the output of this function but we omitted plotting these
        in the example above.

        Returns
        -------
        dict[tuple[int, ...], str]
            Dict mapping data qubits to their pauli charges ("_", "X", "Y", or "Z")
        """
        # Initialize the dictionary of pauli charge numbers
        pauli_charge_numbers = {qb: {"X": 0, "Y": 0, "Z": 0} for qb in self.data_qubits}

        # Count the number of X, Y, and Z numbers for each data qubit in each stabilizer
        for stab in self.stabilizers:
            for i, qb in enumerate(stab.data_qubits):
                pauli_charge_numbers[qb][stab.pauli[i]] += 1

        # Construct the dictionary of pauli charges
        pauli_charges = {
            # Get Pauli charge ("_", "X", "Y", or "Z") for each qubit
            qb: paulixz_to_char(
                (charge["X"] + charge["Y"]) % 2,  # The X charge
                (charge["Z"] + charge["Y"]) % 2,  # The Z charge
                # Note that Y contributes to both X and Z charges
            )
            for qb, charge in pauli_charge_numbers.items()
        }

        return pauli_charges

    @cached_property
    def parity_check_matrix(self) -> ParityCheckMatrix:
        """
        Return the parity check matrix of the Block.

        Returns
        -------
        ParityCheckMatrix
            The parity check matrix of the Block
        """

        # Generate the parity check matrix from the stabilizers
        pc_matrix = ParityCheckMatrix(self.stabilizers)

        return pc_matrix

    @cached_property
    def tanner_graph(self) -> TannerGraph:
        """
        Return the Tanner graph of the Block.

        Returns
        -------
        TannerGraph
            The Tanner graph of the Block
        """

        # Generate the tanner graph from the stabilizers
        tanner_graph = TannerGraph(self.stabilizers)

        return tanner_graph

    # Magic methods
    def __eq__(self, other) -> bool:
        """
        Check whether two blocks are equivalent. The order of stabilizers does not
        matter for the comparison. Also the different fields are checked independently
        on the uuids of objects.

        Returns
        -------
        bool
            True if the two blocks are equivalent, False otherwise
        """

        def check_stabilizer_to_circuit_eq(block1: Block, block2: Block) -> bool:
            """
            Check whether the stabilizer to circuit maps are equivalent. Since the
            stabilizers and syndrome circuits can have different uuids in the two
            blocks respectively, the `stabilizer_to_circuit` dict cannot directly be
            compared. Instead it has to be checked if the dict maps the same elements
            to each other by comparing them element-wise.
            """
            for stab_uuid, circ_uuid in block1.stabilizer_to_circuit.items():
                # Find the stabilizer object in block1
                stab1 = [stab for stab in block1.stabilizers if stab.uuid == stab_uuid][
                    0
                ]
                # Find the corresponding stabilizer object in block2 using the == method
                # which ignores the stabilizer uuid. This is done to check that the two
                # stabilizers are mapped to the same syndrome circuits in both blocks
                # respectively, irrespective of different uuids
                stab2 = [stab for stab in block2.stabilizers if stab == stab1][0]
                # Uuid of the syndrome circuit which stab2 is mapped to
                circ2_uuid = block2.stabilizer_to_circuit[stab2.uuid]
                # Find the two syndrome circuit objects
                circ1 = [
                    circ for circ in block1.syndrome_circuits if circ.uuid == circ_uuid
                ][0]
                circ2 = [
                    circ for circ in block2.syndrome_circuits if circ.uuid == circ2_uuid
                ][0]
                # Compared the two using the __eq__ method which ignores uuids
                if circ1 != circ2:
                    return False

            return True

        if isinstance(other, Block):
            blocks_not_equal = (
                self.unique_label != other.unique_label
                or set(self.stabilizers) != set(other.stabilizers)
                or self.logical_x_operators != other.logical_x_operators
                or self.logical_z_operators != other.logical_z_operators
                or sorted(self.syndrome_circuits, key=lambda x: x.name)
                != sorted(other.syndrome_circuits, key=lambda x: x.name)
                or not check_stabilizer_to_circuit_eq(self, other)
            )
            if blocks_not_equal:
                return False
            # Else, No differences found, the blocks are equivalent
            return True
        # Else, cannot compare the block for equivalence with another object with is
        # not a block
        return NotImplemented

    # Constructors
    @classmethod
    def from_blocks(cls, blocks: tuple[Block, ...]) -> Block:
        """
        Combine multiple blocks into a single block. The blocks must not overlap
        with each other. The output block will be of code type "custom".

        Parameters
        ----------
        blocks : tuple[Block, ...]
            The blocks to be combined.

        Returns
        -------
        Block
            The combined block.
        """
        # Check that the input is indeed an iterable of blocks
        if not all(isinstance(block, Block) for block in blocks):
            raise ValueError("The input argument blocks must only contain blocks.")

        # Check that the blocks do not overlap
        cls.blocks_no_overlap(blocks)

        # Combine the stabilizers and logical operators
        stabilizers = tuple(stab for block in blocks for stab in block.stabilizers)
        logical_x_operators = tuple(
            log_x for block in blocks for log_x in block.logical_x_operators
        )
        logical_z_operators = tuple(
            log_z for block in blocks for log_z in block.logical_z_operators
        )

        # Get syndrome_circuits without duplicates and create a mapping from old to new
        # uuids
        syndrome_circuits_with_duplicates = tuple(
            circ for block in blocks for circ in block.syndrome_circuits
        )
        syndrome_circuits = tuple(set(syndrome_circuits_with_duplicates))

        synd_circ_map_old_to_new_uuid = {
            circ.uuid: syndrome_circuits[syndrome_circuits.index(circ)].uuid
            for circ in syndrome_circuits_with_duplicates
        }

        # Combine the stabilizer to circuit maps
        stabilizer_to_circuit = {
            stab_uuid: synd_circ_map_old_to_new_uuid[circ_uuid]
            for block in blocks
            for stab_uuid, circ_uuid in block.stabilizer_to_circuit.items()
        }

        return cls(
            stabilizers=stabilizers,
            logical_x_operators=logical_x_operators,
            logical_z_operators=logical_z_operators,
            syndrome_circuits=syndrome_circuits,
            stabilizer_to_circuit=stabilizer_to_circuit,
            unique_label="+".join(block.unique_label for block in blocks),
            skip_validation=True,
        )

    # Static methods
    @staticmethod
    def blocks_no_overlap(blocks: tuple[Block, ...]) -> tuple[Block, ...]:
        """
        Check that blocks do no overlap, meaning that no blocks should be defined
        on the same data or ancilla qubits.

        NOTE: We do not check that logical operators are not overlapping. In the `Block`
        class there is a validation checking that logical operators are defined on the
        same qubits as the stabilizers.
        """
        for block_i, block_j in combinations(blocks, 2):
            shared_data_qubits = set(block_i.data_qubits) & set(block_j.data_qubits)
            if len(shared_data_qubits) > 0:
                raise ValueError(
                    f"Block '{block_i.unique_label}' and block "
                    f"'{block_j.unique_label}' share the data qubits "
                    f"{shared_data_qubits}."
                )

            shared_ancilla_qubits = set(block_i.ancilla_qubits) & set(
                block_j.ancilla_qubits
            )
            if len(shared_ancilla_qubits) > 0:
                raise ValueError(
                    f"Block '{block_i.unique_label}' and block "
                    f"'{block_j.unique_label}' share the ancilla qubits "
                    f"{shared_ancilla_qubits}."
                )
        return blocks

    # Methods
    def rename(self, name: str) -> Block:
        """
        Return a copy of the Block with the new name.
        """
        return self.__class__(
            unique_label=name,
            stabilizers=self.stabilizers,
            logical_x_operators=self.logical_x_operators,
            logical_z_operators=self.logical_z_operators,
            syndrome_circuits=self.syndrome_circuits,
            stabilizer_to_circuit=self.stabilizer_to_circuit,
        )

    def shift(
        self,
        position: tuple[int, ...],
        new_label: str | None = None,
    ) -> Block:
        """
        Return a copy of the Block where all qubit coordinates are shifted by a given
        position.

        Parameters
        ----------
        position : tuple[int, ...]
            Vector by which the block should be shifted
        new_label : str | None, optional
            New label for the block. If None, the same label is used.

        Returns
        -------
        Block
            A new Block with the shifted qubit coordinates.
        """

        if (
            len(position) != len(self.logical_x_operators[0].data_qubits[0]) - 1
        ):  # Remove unit vector
            raise ValueError(
                f"The shift position has a wrong dimension of {len(position)}. "
                f"Expected {len(self.logical_x_operators[0].data_qubits[0]) - 1} "
                "instead."
            )

        if new_label is None:
            new_label = self.unique_label

        def shift_coord(
            coord: tuple[int, ...], shift: tuple[int, ...]
        ) -> tuple[int, ...]:
            if coord is None or len(coord) == 0:
                return coord
            return [
                (coord[i] + shift[i] if i < len(shift) else coord[i])
                for i in range(len(coord))
            ]  # Append the old unit vector

        new_stabilizers = [
            Stabilizer(
                pauli=stab.pauli,
                data_qubits=[shift_coord(qb, position) for qb in stab.data_qubits],
                ancilla_qubits=[
                    shift_coord(qb, position) for qb in stab.ancilla_qubits
                ],
            )
            for stab in self.stabilizers
        ]
        logical_x_operators = [
            PauliOperator(
                log_x.pauli,
                data_qubits=[shift_coord(qb, position) for qb in log_x.data_qubits],
            )
            for log_x in self.logical_x_operators
        ]
        logical_z_operators = [
            PauliOperator(
                log_z.pauli,
                data_qubits=[shift_coord(qb, position) for qb in log_z.data_qubits],
            )
            for log_z in self.logical_z_operators
        ]

        # Replacing old stabilizer uuids with new ones
        stabs_map_old_to_new_uuid = {
            stab.uuid: new_stabilizers[i].uuid
            for i, stab in enumerate(self.stabilizers)
        }
        new_stabilizer_to_circuit = {
            stabs_map_old_to_new_uuid[old_stab_uuid]: circ_uuid
            for old_stab_uuid, circ_uuid in self.stabilizer_to_circuit.items()
        }

        return self.__class__(
            unique_label=new_label,
            stabilizers=new_stabilizers,
            logical_x_operators=logical_x_operators,
            logical_z_operators=logical_z_operators,
            syndrome_circuits=self.syndrome_circuits,
            stabilizer_to_circuit=new_stabilizer_to_circuit,
            skip_validation=True,  # Skip validation since the block has been validated
        )

    @cached_property
    def stabilizers_labels(self) -> dict[str, dict[str, tuple[int, ...]]]:
        """
        Builds a dictionary associating stabilizers, via their uuid, with a
        set of labels defined through a dictionary. Inside Block, this is generically
        populated with the space coordinates of the stabilizer check, corresponding to
        the ancilla which measures each stabilizer.

        This functionality can be leveraged to later provide these labels to Syndromes
        and Detectors associated with a given Stabilizer.

        Returns
        -------
        dict[str, dict[str, tuple[int, ...]]]
            Dictionary associating stabilizer uuids with their labels.
        """

        # Retrieve space coordinates
        labels = {
            stab.uuid: {"space_coordinates": stab.ancilla_qubits[0]}
            for stab in self.stabilizers
        }

        return labels

    def get_stabilizer_label(self, stabilizer_uuid: str) -> dict[str, tuple[int, ...]]:
        """
        Get the labels of a stabilizer, specified by its uuid.

        Parameters
        ----------
        stabilizer_uuid : str
            uuid of the stabilizer.

        Returns
        -------
        dict[str, tuple[int, ...]]
            Labels of the stabilizer.
        """

        return self.stabilizers_labels.get(stabilizer_uuid, {})



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from .block import Block
from .channel import Channel, ChannelType
from .circuit import Circuit
from .ifelse_circuit import IfElseCircuit
from .circuit_algorithms import (
    coloration_circuit,
    cardinal_circuit,
    generate_stabilizer_and_syndrome_circuits_from_algorithm,
    extract_syndrome_circuit,
)
from .eka import Eka
from .lattice import Lattice, LatticeType
from .logical_state import LogicalState
from .matrices import ClassicalParityCheckMatrix, ParityCheckMatrix
from .pauli_operator import PauliOperator
from .stabilizer import Stabilizer
from .syndrome_circuit import SyndromeCircuit
from .tanner_graphs import (
    TannerGraph,
    ClassicalTannerGraph,
    cartesian_product_tanner_graphs,
    verify_css_code_stabilizers,
)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import numpy as np

from .graph_matrix_utils import binary_gaussian_elimination
from .pauli_binary_vector_rep import SignedPauliOp, pauliops_anti_commute


def tableau_generates_pauli_group(tableau: np.ndarray) -> bool:
    """
    Given a tableau, check if all of its operators, (stabilizers & destabilizers) can
    generate the pauli group. This is done by checking if the tableau is equivalent to
    the identity matrix after performing gaussian elimination.

    Parameters
    ----------
    tableau : np.ndarray
        The tableau to be checked.

    Returns
    -------
    bool
        Whether the tableau generates the Pauli group.
    """

    _, ncols = tableau.shape
    nqubits = ncols // 2

    # remove signs (last column) from the tableau, if present
    if ncols % 2 == 1:
        tableau = tableau[:, : ncols - 1]

    # perform gaussian elimination
    post_gauss_tab = binary_gaussian_elimination(tableau)

    # initialize identity matrix to check pauli group generation
    id_matr = np.eye(2 * nqubits, dtype=post_gauss_tab.dtype)

    return np.array_equal(post_gauss_tab, id_matr)


def is_tableau_valid(tableau: np.ndarray) -> bool:
    """
    Checks if a tableau is valid. This is done by checking if it generates the pauli
    group and whether the correct commutation relations hold.

    For the commutation relations, the tableau is split into two sets of
    stabilizers of equal size calling them S (stabilizers) and D (destabilizers). The
    following is then checked:

    * All stabilizers of S commute with each other and all stabilizers of D
        commute with each other.
    * Stabilizer i from list S has to anti-commute with stabilizer j from list
        D if i=j. If i != j, they have to commute.

    Parameters
    ----------
    tableau : np.ndarray
        The tableau to be checked.

    Returns
    -------
    bool
        Whether the tableau is valid.
    """

    nrows, _ = tableau.shape
    nstabs = nrows // 2

    # check if it generates the pauli group
    # NOTE: It might be the case that this test is not needed, i.e. if the other
    # conditions are fulfilled, this may be certain.
    generates_pauli_group = tableau_generates_pauli_group(tableau)

    if not generates_pauli_group:
        return False

    # Check if commutation relations are violated
    # We are comparing pairs of operators by calculating
    # the 2d array that shows the anti-commutation of
    # pairs stabilizers/destabilizers.
    destab_array = tableau[:nstabs, :]
    stab_array = tableau[nstabs:, :]

    # all commutations of stabilizers should be zero
    # [s_i, s_j] = 0
    if any(
        pauliops_anti_commute(
            SignedPauliOp(stab_array[s1_idx, :]), SignedPauliOp(stab_array[s2_idx, :])
        )
        for s1_idx in range(nstabs)
        for s2_idx in range(s1_idx + 1)
    ):
        return False

    # all commutations of destabilizers should be zero
    # [d_i, d_j] = 0
    if any(
        pauliops_anti_commute(
            SignedPauliOp(destab_array[d1_idx, :]),
            SignedPauliOp(destab_array[d2_idx, :]),
        )
        for d1_idx in range(nstabs)
        for d2_idx in range(d1_idx, nstabs)
    ):
        return False

    # Commutation relations for stabilizers-destabilizers:
    # {s_i, d_j} = 0 iff i=j
    if any(
        pauliops_anti_commute(
            SignedPauliOp(stab_array[s_idx, :]), SignedPauliOp(destab_array[d_idx, :])
        )
        ^ (s_idx == d_idx)
        for s_idx in range(nstabs)
        for d_idx in range(s_idx, nstabs)
    ):  # iff i=j, then flip the condition
        return False

    return True



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka.operations.code_operation import (
    ConditionalLogicalX,
    ConditionalLogicalY,
    ConditionalLogicalZ,
)
from ..interpretation_step import InterpretationStep


# pylint: disable=unused-argument, duplicate-code
def conditional_logical_pauli(
    interpretation_step: InterpretationStep,
    operation: ConditionalLogicalX | ConditionalLogicalY | ConditionalLogicalZ,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Apply a conditional logical X, Y or Z operator.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step to which the operation should be applied.
    operation : ConditionalLogicalX | ConditionalLogicalY | ConditionalLogicalZ
        The operation to be applied, can either be a conditional logical X, Y or Z
        operation.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block.

    Returns
    -------
    InterpretationStep
        New InterpretationStep containing all modifications due to the logical pauli
        gate.
    """

    block = interpretation_step.get_block(operation.input_block_name)
    logical_qubit = operation.logical_qubit

    # Check if the logical qubit exists
    if logical_qubit >= block.n_logical_qubits:
        raise ValueError(
            f"Logical qubit {logical_qubit} does not exist in block "
            f"{block.unique_label}"
        )

    # Conditionally applying LogicalZ (LogicalX) is the same as adding that condition to
    # logical_x_operator_updates (logical_z_operator_updates).
    # Conditionally applying LogicalY is the same as adding that condition to both
    # logical_x_operator_updates and logical_z_operator_updates.

    logical_update = interpretation_step.logical_measurements.get(operation.condition)
    if logical_update is None:
        raise ValueError(
            f"Logical Measurement {operation.condition} has not yet been made"
        )
    match operation.__class__.__name__:
        case "ConditionalLogicalX":
            interpretation_step.update_logical_operator_updates_MUT(
                "Z",
                block.logical_z_operators[logical_qubit].uuid,
                logical_update,
                inherit_updates=False,
            )
            return interpretation_step
        case "ConditionalLogicalZ":
            interpretation_step.update_logical_operator_updates_MUT(
                "X",
                block.logical_x_operators[logical_qubit].uuid,
                logical_update,
                inherit_updates=False,
            )
        case "ConditionalLogicalY":
            interpretation_step.update_logical_operator_updates_MUT(
                "X",
                block.logical_x_operators[logical_qubit].uuid,
                logical_update,
                inherit_updates=False,
            )
            interpretation_step.update_logical_operator_updates_MUT(
                "Z",
                block.logical_z_operators[logical_qubit].uuid,
                logical_update,
                inherit_updates=False,
            )
        case _:
            raise ValueError(f"Operation {operation.__class__.__name__} not supported")

    return interpretation_step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Eka
from loom.eka.operations import Operation

from ..interpretation_step import InterpretationStep


class BaseApplicator:  # pylint: disable=too-few-public-methods
    """
    Base Class for Applicators. BaseOperations are implemented at the level of this
    BaseApplicator. For CodeOperations and LogicalOperations, subclasses are used.

    Classmethods of this class evolve the InterpretationStep object, by applying the
    operation to it. The InterpretationStep object is then returned.
    The signature of applicator methods should be:
    def operation_name(self, interpretation_step: InterpretationStep,
    operation: Operation, same_timeslice: bool, debug_mode: bool) -> InterpretationStep:
    """

    def __init__(
        self,
        eka: Eka,
    ):
        # Define supported base operations
        self.supported_operations = {}
        self.eka = eka

    def apply(
        self,
        interpretation_step: InterpretationStep,
        operation: Operation,
        same_timeslice: bool,
        debug_mode: bool,
    ) -> InterpretationStep:
        """
        If the `operation` can be found in the `supported_operations` dictionary,
        this method applies the operation to the input `interpretation_step`.

        The class name of the operation is used as a way to determine
        if the operation is supported by this applicator.

        Parameters
        ----------
        interpretation_step : InterpretationStep
            Interpretation step containing the blocks to be modified.
        operation : :class:`loom.eka.operations.base_operation.Operation`
            The operation to be applied. It should be an instance of a class
            that is registered in `supported_operations`.
        same_timeslice : bool
            Flag indicating whether the operation is part of the same timestep as the
            previous operation.
        debug_mode : bool
            Flag indicating whether the interpretation should be done in debug mode.
            Currently, the effects of debug mode are:
            - Disabling the commutation validation of Block

        Returns
        -------
        InterpretationStep
            Interpretation step after the operation is applied.
        """
        operation_function = self.supported_operations.get(operation.__class__.__name__)
        if operation_function is None:
            raise NotImplementedError(
                f"Operation {operation.__class__.__name__} is not supported by "
                f"{self.__class__.__name__}"
            )
        return operation_function(
            interpretation_step, operation, same_timeslice, debug_mode
        )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from functools import reduce
from pydantic.dataclasses import dataclass
import numpy as np

from ..eka import Stabilizer, Block
from ..eka.utilities import SignedPauliOp
from ..cliffordsim import Engine
from ..cliffordsim.operations import UpdateTableau, Operation


from .check_abstract import AbstractValidityCheck
from .utilities import get_all_cliffordsim_registers_with_random_flags


@dataclass(frozen=True)
class StabilizerMeasurementCheckOutput:
    """
    Dataclass to store the output of the Stabilizer Measurement check.

    Parameters
    ----------
    expected_vs_measured_stabs: tuple[tuple[str,Stabilizer,tuple[Stabilizer,...]],...]
        A tuple containing the stabilizers that were not measured correctly. Each tuple
        contains the measurement index, the stabilizer that was supposed to be measured
        by it, and the stabilizers that were measured instead.
    probabilistic_measurements: tuple[str, ...]
        A tuple containing all the measurements that are probabilistic
    """

    expected_vs_measured_stabs: tuple[
        tuple[str, Stabilizer, tuple[Stabilizer, ...]], ...
    ]
    probabilistic_measurements: tuple[str, ...]

    def __len__(self):
        return len(self.expected_vs_measured_stabs) + len(
            self.probabilistic_measurements
        )

    def __str__(self):
        out = ""
        if any(self.expected_vs_measured_stabs):
            out = "- Expected vs Measured Stabilizers:\n"
            for idx, stab, measured_stabs in self.expected_vs_measured_stabs:
                measured_stabs_str = (
                    ", ".join(str(stab) for stab in measured_stabs)
                    if measured_stabs
                    else "None"
                )
                out += (
                    f"Measurement {idx}: Expected {stab}, "
                    f"Measured: {measured_stabs_str}\n"
                )
        if any(self.probabilistic_measurements):
            out += "- Probabilistic Measurements: "
            out += ", ".join(str(idx) for idx in self.probabilistic_measurements)
            out += "\n"
        return out.rstrip("\n")  # Remove the last newline character for cleaner output


@dataclass(frozen=True)
class StabilizerMeasurementCheck(AbstractValidityCheck):
    """Dataclass to store the results of the Stabilizer Measurement check.

    Parameters
    ----------
    output: StabilizerMeasurementCheckOutput
        An object containing the output of the check.

    Properties
    -----------
    message: str
        A message indicating the result of the check. It will be empty if the check is
        valid, otherwise it will contain a message describing the issue.
    valid: bool
        True if the check is valid (i.e., all measurements were deterministic and
        measured the assigned stabilizer), False otherwise.
    """

    output: StabilizerMeasurementCheckOutput

    @property
    def message(self) -> str:
        match (
            len(self.output.expected_vs_measured_stabs),
            len(self.output.probabilistic_measurements),
        ):
            case (0, 0):
                return ""
            case (0, _):
                return "Some measurement(s) were not deterministic."
            case (_, 0):
                return "Some measurement(s) did not measure the assigned stabilizer."
            case (_, _):
                return (
                    "Some measurement(s) were not deterministic and some did not "
                    "measure the assigned stabilizer."
                )


# pylint: disable=too-many-branches, line-too-long, anomalous-backslash-in-string
def check_input_stabilizer_measurement(
    base_cliffordsim_operations: tuple[Operation, ...],
    input_block: Block,
    measurement_to_stabilizer_map: dict[str, Stabilizer],
    seed: int | None,
) -> StabilizerMeasurementCheck:
    """Checks whether the measurement of the stabilizers happened and whether the
    result was registered at the correct place. This is done by running the circuit with
    an initial state of :math:`\ket{00...0}` and recording the measurement results.
    Then, for each stabilizer, the stabilizer is flipped and the circuit is run again.
    The measurement results are compared to the reference results to check if the
    measurement was correct. The stabilizers that were not measured correctly are stored
    in the output.

    Parameters
    ----------
    base_cliffordsim_operations: tuple[:class:`loom.cliffordsim.operations.base_operation.Operation`, ...]
        The cliffordsim operations that represent the circuit to be checked.
    input_block: Block
        The Block object that represents the input code.
    measurement_to_input_stabilizer_map: dict[str, Stabilizer]
        Dictionary matching the classical channel name of a measurement operation with a
        stabilizer in the input code.
    seed: int | None
        The seed for the cliffordsim engine.

    Returns
    -------
    StabilizerMeasurementCheck
        The result of the Stabilizer Measurement check.
    """

    # Get the measurement results from the reference run
    mresults_reference = get_mresults_with_flipped_reduced_stabilizers(
        [], base_cliffordsim_operations, input_block, seed
    )
    # Check if the measurement indices given by the user are valid
    for chan_label in measurement_to_stabilizer_map.keys():
        if chan_label not in mresults_reference:
            raise ValueError(
                f"Measurement channel label {chan_label} not found in cliffordsim data store."
            )

    # Get the deterministic and probabilistic measurements
    probabilistic_measurements: set[str] = set()
    measurement_reference: dict[str, int] = {}
    for chan_label, (meas_result, meas_is_random) in mresults_reference.items():
        if chan_label in measurement_to_stabilizer_map:
            if meas_is_random:
                probabilistic_measurements.add(chan_label)
            else:
                measurement_reference[chan_label] = meas_result

    # For every measurement find the stabilizers that contributed to the measurement
    contributions_present: dict[str, set[Stabilizer]] = {
        i: set() for i in measurement_reference
    }
    # Check what happens by flipping one by one the stabilizers
    for stab_flipped_idx, stab_flipped in enumerate(input_block.stabilizers):
        # Use bookkeeping to find which reduced stabilizers to flip.
        # The rows that need to be flipped are the ones that have the value True in the
        # column stab_flipped_idx
        reduced_stab_idx_to_flip = input_block.reduced_bookkeeping[:, stab_flipped_idx]

        if not np.any(reduced_stab_idx_to_flip):
            # If the stabilizer does not appear in the reduced stabilizer array, skip
            # as it will be the same as running the reference we did previously
            # The stabilizer is not present in the reduced stabilizer array because in
            # overdefined sets of stabilizers, not all stabilizers are needed to
            # create the reduced stabilizer array.
            continue

        # Get the measurement results with the flipped stabilizers
        mresults_flipped = get_mresults_with_flipped_reduced_stabilizers(
            reduced_stab_idx_to_flip, base_cliffordsim_operations, input_block, seed
        )

        # Go through the measurement results to check for consistency and the stabilizer
        # contributions
        for chan_label, (meas_result, meas_is_random) in mresults_flipped.items():

            if chan_label not in measurement_to_stabilizer_map:
                # Skip if the measurement is not in the map
                continue

            # Check for any inconsistencies in the measurements being deterministic
            measurement_should_be_random = chan_label in probabilistic_measurements
            if measurement_should_be_random != meas_is_random:
                raise RuntimeError(
                    f"Measurement with index {chan_label} is sometimes random and "
                    "sometimes deterministic."
                )

            if meas_is_random:
                # Skip if the measurement is random
                continue

            # Check that the measurement result matches the sign of the stabilizer
            # in the initialized array
            if meas_result != measurement_reference[chan_label]:
                contributions_present[chan_label].add(stab_flipped)

    n_dqubits = input_block.n_data_qubits
    # Gather the stabilizers that were not measured correctly
    expected_vs_measured_stabs = tuple(
        (idx, stab, contributions_present[idx])
        for idx, stab in measurement_to_stabilizer_map.items()
        # Do not include the deterministic measurements
        if idx not in probabilistic_measurements
        # Do not include the measurements that indeed measured the stabilizer or the
        # product of the stabilizers that it measured is equivalent to the correct
        # stabilizer
        and not reduce(
            lambda x, y: x * y,
            (
                stab.as_signed_pauli_op(input_block.data_qubits)
                for stab in contributions_present[idx]
            ),
            SignedPauliOp.identity(n_dqubits),
        )  # The product of the measured stabilizers
        == stab.as_signed_pauli_op(input_block.data_qubits)  # The correct stabilizer
    )

    # Sort based on the first element of the tuple
    expected_vs_measured_stabs = tuple(
        sorted(expected_vs_measured_stabs, key=lambda x: x[0])
    )

    # Initialize the output dataclass
    stab_meas_check_output = StabilizerMeasurementCheckOutput(
        expected_vs_measured_stabs, sorted(tuple(probabilistic_measurements))
    )

    return StabilizerMeasurementCheck(stab_meas_check_output)


# pylint: disable=anomalous-backslash-in-string
def get_mresults_with_flipped_reduced_stabilizers(
    reduced_stab_idx_to_flip: np.ndarray | list[int],
    base_cliffordsim_operations: tuple[Operation, ...],
    input_block: Block,
    seed: int | None,
) -> dict[str, tuple[int, bool | None]]:
    """
    This function runs the circuit with some stabilizers flipped and returns the
    measurement results. The logical state that is used is the :math:`\ket{00...0}` state.

    Parameters
    ----------
    reduced_stab_idx_to_flip : np.ndarray | list[int]
        The indices of the reduced stabilizers to flip.
    base_cliffordsim_operations : tuple[ \
        :class:`loom.cliffordsim.operations.base_operation.Operation`, ...]
        The base cliffordsim operations that will be used to run the circuit.
    input_block : Block
        The Block object that represents the input code.
    seed : int | None
        The seed for the cliffordsim engine.

    Returns
    -------
    dict[str, tuple[int, bool | None]]
        A dictionary containing the names of the classical registers and their values,
        along with a boolean indicating whether the value is a result of a random
        measurement. If the value is not a result of a measurement, the boolean is None.
    """
    stab_flipped_base_array = input_block.reduced_stabarray.array.copy()
    # Flip the signs of the stabilizers that need to be flipped
    stab_flipped_base_array[reduced_stab_idx_to_flip, -1] ^= 1
    # Construct tableau corresponding to the array with flipped stabilizer
    # while using logical state |00...0>
    stab_flipped_tableau = np.vstack(
        (
            input_block.x_log_stabarray.array,
            input_block.destabarray.array,
            input_block.z_log_stabarray.array,
            stab_flipped_base_array,
        )
    )

    # Run the circuit from the initial state corresponding to the tableau
    # with the flipped stabilizers
    operations = (UpdateTableau(stab_flipped_tableau), *base_cliffordsim_operations)
    cliffordsim_engine = Engine(operations, input_block.n_data_qubits, seed=seed)
    cliffordsim_engine.run()

    return get_all_cliffordsim_registers_with_random_flags(cliffordsim_engine)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Stabilizer, Circuit
from loom.eka.utilities import Direction, DiagonalDirection, Orientation
from loom.interpreter import InterpretationStep, Cbit
from loom.interpreter.applicator import generate_syndromes, generate_detectors

from loom_rotated_surface_code.code_factory import RotatedSurfaceCode
from loom_rotated_surface_code.operations import MoveBlock
from .utilities import (
    shift_block_towards_direction,
    find_detailed_schedules,
    direction_to_coord,
    DetailedSchedule,
    update_qubit_coords,
)


def move_block(
    interpretation_step: InterpretationStep,
    operation: MoveBlock,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Applicator to move a block in a fault tolerant manner based on the MoveBlock
    operation. The Block will be moved 1-unit in the specified direction by moving
    twice diagonally. By default, the diagonal moves are chosen using a secondary
    direction TOP or LEFT.

    For example, if the direction is RIGHT, the block will move diagonally upwards to
    the right in the first syndrome extraction round and diagonally downwards to the
    right in the second round.

    The algorithm is as follows:

    - A.) Begin MoveBlock composite operation session

    - B.) Valid move check
        - B.1) Check if the qubits required for the block to be moved are available.

    - C.) Shift the block
        - C.1) Shift the block in the specified direction
        - C.2) Update all the evolutions
        - C.3) Propagate all the updates

    - D.) Circuit generation
        - D.1) Find the qubit initializations required for the swap-then-qec operation \
              and create the reset circuit
        - D.2) Find the stabilizer schedules for the new block and generate the cnot \
              circuit
        - D.3) Generate the teleportation finalization circuit with necessary updates \
              to stabilizers and logical operators
        - D.4) Final measurement of the stabilizers and generation of syndromes
        - D.5) Combine all the circuits into one diagonal move circuit
    
    - E.) Append necessary information to the interpretation step
        - E.1) Append the circuit to the interpretation step
        - E.2) Update the block history and evolution
        - E.3) Create and append the new syndromes and detectors

    - Repeat steps B, C, D, E for the second diagonal direction.

    - F.) Final Circuit
        - F.1) End the composite operation session and append the full move block \
        circuit


    If the block is moved to the top:

    The syndrome extraction circuits assign the qubits as follows:

    1. The first set of syndrome extraction rounds assigns the qubits as follows::

                    a
            x --- x --- x
         a  |  a  |  a  |
            x --- x --- x
            |  a  |  a  |  a
            x --- x --- x
                a


        where x are data channels and a are ancilla channels.

    2. The second set of syndrome extraction rounds creates a new set of channels
    BASED on how the channels are first created, in most cases as ancilla qubits of
    either the teleport circuits or that of the syndrome extraction rounds during the
    move.

    As such the new set of qubits are assigned as follows::

            a     a     a
               x --- x --- x
            a  |  a  |  a  |
               x --- x --- x
            a  |  a  |  a  |  a
               x --- x --- x
                  a


    2b. Following step::

               x     x     x
            a --- a --- a
         x  |  x  |  x  |  x
            a --- a --- a
            |  x  |  x  |  x
            a --- a --- a     a
               x     x     x
                  a


    2c. Final Expected Configuration::

                        a
               x --- x --- x
            a  |  a  |  a  |
         a     x --- x --- x
            a  |  a  |  a  |  a
               x --- x --- x
            a     a     a     a
               x     x     x
                  a

    Parameters
    ----------
    interpretation_step: InterpretationStep
        The current interpretation step.
    operation: MoveBlock
        MoveBlock operation description.

    Returns
    -------
    InterpretationStep
        Interpretation Step after the MoveBlock operation has been applied.

    Raises
    ------
    ValueError
        If the block is not 2D or if the block cannot be moved in the specified
        direction.
    """
    block = interpretation_step.get_block(operation.input_block_name)

    # A) Begin MoveBlock composite operation session
    interpretation_step.begin_composite_operation_session_MUT(
        same_timeslice=same_timeslice,
        circuit_name=(
            f"Move block {block.unique_label} towards {operation.direction.value}"
        ),
    )

    # Find occupied qubits from other blocks in the latest timeslice
    other_blocks = [
        each_block
        for each_block in interpretation_step.get_blocks_at_index(-1)
        if each_block != block
    ]
    occupied_qubits = [
        each_qubit for each_block in other_blocks for each_qubit in each_block.qubits
    ]

    # Decompose the direction into 2 diagonal directions
    decomposed_directions = composite_direction(operation.direction)

    current_block = block
    for diag_direction in decomposed_directions:
        # B) - Valid move check
        # B.1) Check if the qubits required for the block to be moved are available.
        check_valid_move(occupied_qubits, current_block.qubits, diag_direction)

        # C, D, E) - Move the block diagonally via swap-QEC
        interpretation_step = move_block_diagonally_via_swap_qec(
            interpretation_step, current_block, diag_direction, debug_mode
        )

        # Prepare for the next iteration
        current_block = interpretation_step.get_block(operation.input_block_name)

    # F) Final Circuit
    # F.1) End the composite operation session and append the full move block circuit
    move_block_circuit = interpretation_step.end_composite_operation_session_MUT()
    interpretation_step.append_circuit_MUT(move_block_circuit, same_timeslice)

    return interpretation_step


def move_block_diagonally_via_swap_qec(
    interpretation_step: InterpretationStep,
    current_block: RotatedSurfaceCode,
    diag_direction: Direction,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Moves the block diagonally in the specified direction using a swap-then-qec
    procedure.

    Parameters
    ----------
    interpretation_step: InterpretationStep
        The current interpretation step.
    current_block: RotatedSurfaceCode
        The current block to be moved.
    diag_direction: Direction
        The diagonal direction in which the block is to be moved.
    debug_mode: bool
        If True, skip validation when creating the new block.
    Returns
    -------
    tuple[RotatedSurfaceCode, Circuit]
        The new block after the move and the circuit that performs the move.
    """

    # C) - Shift the block
    # C.1) Shift the block in the specified direction
    new_block, stab_ev, logx_ev, logz_ev = shift_block_towards_direction(
        current_block, diag_direction, debug_mode=debug_mode
    )

    # C.2) Update all the evolutions
    # Stabilizer evolution
    interpretation_step.stabilizer_evolution.update(stab_ev)
    # Logical operator evolution
    interpretation_step.logical_x_evolution.update(logx_ev)
    interpretation_step.logical_z_evolution.update(logz_ev)

    # C.3) Propagate all the updates
    # Stabilizer updates
    for new_stab_uuid, old_stab_uuids in stab_ev.items():
        propagated_updates = ()
        for old_stab_uuid in old_stab_uuids:
            propagated_updates += interpretation_step.stabilizer_updates.get(
                old_stab_uuid, ()
            )
        if propagated_updates:
            interpretation_step.stabilizer_updates[new_stab_uuid] = propagated_updates
    # Logical operator updates
    interpretation_step.update_logical_operator_updates_MUT(
        "X", new_block.logical_x_operators[0].uuid, (), True
    )
    interpretation_step.update_logical_operator_updates_MUT(
        "Z", new_block.logical_z_operators[0].uuid, (), True
    )

    # D) Circuit generation
    # D.1) Find the qubit initializations required for the swap-then-qec operation
    # and create the reset circuit
    (
        anc_qubits_to_init,
        data_qubits_to_init,
        teleportation_qubit_pairs,
    ) = find_swap_then_qec_qubit_initializations(
        current_block.stabilizers, diag_direction
    )
    reset_basis_circuit = Circuit(
        name=("Initialization of qubits for first swap-then-qec"),
        circuit=[
            [
                Circuit(
                    f"reset_{'0' if pauli == 'Z' else '+'}",
                    channels=[interpretation_step.get_channel_MUT(q)],
                )
                for pauli in ["X", "Z"]
                for q in anc_qubits_to_init[pauli] + data_qubits_to_init[pauli]
            ]
        ],
    )

    # D.2) Find the stabilizer schedules for the new block and generate the cnot
    # circuit
    stab_schedule_dict = find_detailed_schedules(new_block, diag_direction)
    cnots_circuit = get_swap_qec_cnots(
        interpretation_step,
        new_block,
        diag_direction,
        stab_schedule_dict,
        anc_qubits_to_init,
        data_qubits_to_init,
    )

    # D.3) Generate the teleportation finalization circuit with necessary updates
    # to stabilizers and logical operators
    tp_finalization = generate_teleportation_measurement_circuit_with_updates(
        interpretation_step,
        new_block,
        anc_qubits_to_init,
        teleportation_qubit_pairs,
    )

    # D.4) Final measurement of the stabilizers and generation of syndromes
    syndrome_meas_circ, stab_measurements = (
        generate_syndrome_measurement_circuit_and_cbits(
            interpretation_step,
            new_block,
        )
    )

    # D.5) Combine all the circuits into one diagonal move circuit
    circ = Circuit(
        name=f"move block {new_block.unique_label} towards {diag_direction.value}",
        circuit=Circuit.construct_padded_circuit_time_sequence(
            [
                (reset_basis_circuit,),
                (cnots_circuit,),
                (tp_finalization, syndrome_meas_circ),
            ]
        ),
    )

    # E) - Append necessary information to the interpretation step
    # E.1) Append the circuit to the interpretation step
    interpretation_step.append_circuit_MUT(circ, same_timeslice=False)

    # E.2) Update the block history and evolution
    interpretation_step.update_block_history_and_evolution_MUT(
        (new_block,), (current_block,)
    )

    # E.3) Create and append the new syndromes and detectors
    # Create all new syndromes
    generate_and_append_block_syndromes_and_detectors(
        interpretation_step=interpretation_step,
        block=new_block,
        syndrome_measurement_cbits=stab_measurements,
    )

    return interpretation_step


def check_valid_move(
    occupied_qubits: tuple[tuple[int, int, int], ...],
    moving_qubits: tuple[tuple[int, int, int], ...],
    direction: Direction | DiagonalDirection,
):
    """
    This function checks if the move operation is valid by ensuring that none of the
    moving qubits will be moved onto an occupied qubit. If any of the moving qubits
    will be moved onto an occupied qubit, a ValueError is raised.

    Parameters
    ----------
    occupied_qubits: tuple[tuple[int, int, int], ...]
        A tuple of qubit coordinates representing "occupied" qubits. Qubits taking part
        in the move, cannot be moved onto these set of qubits as they are "occupied".
    moving_qubits: tuple[tuple[int, int, int], ...]
        A tuple of qubit coordinates representing qubits that will be moved. The last
        value in each qubit coordinate tuple represents the sub-lattice index of the
        qubit.
    direction: Direction | DiagonalDirection
        The direction in which the qubits will be moving.

    Raises
    ------
    ValueError
        If the move is invalid, i.e., if any of the moving qubits will be moved onto an
        occupied qubit.
    """
    # Do the diagonal move.
    for each_qubit in moving_qubits:
        new_qubit_coords = tuple(
            q + dir
            for q, dir in zip(
                each_qubit,
                direction_to_coord(direction, each_qubit[-1]),
                strict=True,
            )
        )
        if new_qubit_coords in occupied_qubits:
            raise ValueError(
                f"The move operation is invalid. The following qubit, {each_qubit}, is "
                f"moving to an occupied qubit, {new_qubit_coords}."
            )


def composite_direction(
    direction: Direction,
) -> tuple[DiagonalDirection, DiagonalDirection]:
    """
    For the move operation, the user defines only "right", "left", "top" or "bottom".
    However, the actual movement of the block is done via 2 diagonal directions. This
    function returns a tuple of 2 sets of directions based on the user-defined
    direction.

    For e.g.
    If the user specifies "right", the actual movement involves qubits moving to the
    "top right" then "bottom right".
    It would return a tuple with a set of directions "right" and "bottom", and a tuple
    with the directions "top" and "right", representing "bottom right" and "top right"
    respectively.

    composite_direction(Direction.RIGHT)
    -> (DiagonalDirection.TOP_RIGHT, DiagonalDirection.BOTTOM_RIGHT)

    Parameters
    ----------
    direction: Direction
        The direction in which the block is to be moved.

    Returns
    -------
    tuple[DiagonalDirection, DiagonalDirection]
        A tuple of 2 sets of diagonal directions which together represent the composite
        direction of movement.
    """

    composite_directions = {
        Direction.TOP: (
            DiagonalDirection.TOP_LEFT,
            DiagonalDirection.TOP_RIGHT,
        ),
        Direction.RIGHT: (
            DiagonalDirection.TOP_RIGHT,
            DiagonalDirection.BOTTOM_RIGHT,
        ),
        Direction.LEFT: (
            DiagonalDirection.TOP_LEFT,
            DiagonalDirection.BOTTOM_LEFT,
        ),
        Direction.BOTTOM: (
            DiagonalDirection.BOTTOM_LEFT,
            DiagonalDirection.BOTTOM_RIGHT,
        ),
    }

    return composite_directions[direction]


def find_swap_then_qec_qubit_initializations(  # pylint: disable=too-many-locals
    stabilizers: list[Stabilizer],
    diag_direction: DiagonalDirection,
    relocation_diag_direction: DiagonalDirection = None,
) -> tuple[
    dict[str, list[tuple[int, ...]]],
    dict[str, list[tuple[int, ...]]],
    list[tuple[tuple[int, ...], tuple[int, ...]]],
]:
    """
    Find the qubits to initialize for the swap_then_qec operation in the y_wall_out
    operation context. This is for a subset of stabilizers of the block that are
    associated with the boundary qubits.

    The recipe is as follows for moving towards TOP-RIGHT:

    - 1) FIND (non-teleporting) ANCILLA QUBIT INITIALIZATIONS:

        For each stabilizer that is NOT a the TOP-RIGHT boundary stabilizer,
        find the ancilla qubit on the TOP-RIGHT of the initial ancilla qubit
        and initialize it in the basis corresponding to the stabilizer.

    - 2) FIND TELEPORTATION QUBIT PAIRS:

        For each stabilizer that is NOT a BOTTOM-LEFT boundary stabilizer,
        check if the ancilla qubit of the stabilizer is already initialized. If not,
        initialize it in the basis corresponding to the stabilizer and form a
        teleportation qubit pair with the data qubit on the BOTTOM-LEFT of the ancilla.

    - 3) FIND DATA QUBIT INITIALIZATIONS:

        For each stabilizer that is a TOP-RIGHT boundary stabilizer, find the data qubit
        on the TOP-RIGHT of the ancilla qubit and initialize it in the basis
        corresponding to the stabilizer.

    Say we have the following block annotated by its stabilizers::

                     Z
            o --- o --- o
         X  |  Z  |  X  |
            o --- o --- o
            |  X  |  Z  |  X
            o --- o --- o
               Z

    To perform SWAP-THEN-QEC to move the block diagonally towards the TOP-RIGHT, we
    need to initialize the qubits  in the basis as shown below::

                    --- z
               x  |  z  |
            o --- o --- o
            |  z  |  x  |  z
            o --- o --- o --- x
            |  x  |  z  |  x
            o --- o --- o

    Of these, the teleportation qubit pairs can be seen with the numberings below::

                    --- z
               x  |  z  |
            o --- o --- o
            |  1  |  x  |  z
            1 --- o --- o --- x
            |  2  |  z  |  3
            2 --- o --- 3


    Parameters
    ----------
    stabilizers: list[Stabilizer]
        The stabilizers of the block.
    diag_direction: DiagonalDirection
        The diagonal direction that the block is moving.
    relocation_diag_direction: DiagonalDirection
        The directions to relocate the qubits. If None, the qubits are not relocated.

    Returns
    -------
    dict[str, list[tuple[int, ...]]]
        The ancilla qubits to initialize in the X and Z basis.
    dict[str, list[tuple[int, ...]]]
        The data qubits to initialize in the X and Z basis.
    list[tuple[tuple[int, ...], tuple[int, ...]]]
        The teleportation qubit pairs. The first qubit is the ancilla qubit and the
        second is the data qubit.
    """

    # Extract the component directions from the diagonal direction
    vert_direction = diag_direction.direction_along_orientation(Orientation.VERTICAL)
    hor_direction = diag_direction.direction_along_orientation(Orientation.HORIZONTAL)

    # Find the data qubits
    data_qubits = list({q for stab in stabilizers for q in stab.data_qubits})

    data_qubit_sublattice_index = set(q[2] for q in data_qubits)
    if len(data_qubit_sublattice_index) != 1:
        raise ValueError(
            "The stabilizers should all belong to the same sublattice of data qubits."
        )
    data_qubit_sublattice_index = data_qubit_sublattice_index.pop()
    anc_qubit_sublattice_index = 1 - data_qubit_sublattice_index

    # Find the boundary qubits
    bound_coords = {
        Direction.LEFT: min(q[0] for q in data_qubits),
        Direction.RIGHT: max(q[0] for q in data_qubits),
        Direction.TOP: min(q[1] for q in data_qubits),
        Direction.BOTTOM: max(q[1] for q in data_qubits),
    }
    # Find the boundary qubits towards the mixed direction and the opposite direction
    same_direction_boundary_qubits = [
        q
        for q in data_qubits
        if q[0] == bound_coords[hor_direction] or q[1] == bound_coords[vert_direction]
    ]
    opposite_direction_boundary_qubits = [
        q
        for q in data_qubits
        if q[0] == bound_coords[hor_direction.opposite()]
        or q[1] == bound_coords[vert_direction.opposite()]
    ]

    # Find the stabilizers that are only associated with the boundary qubits
    mixed_direction_bound_stabs = [
        stab
        for stab in stabilizers
        if set(stab.data_qubits).issubset(same_direction_boundary_qubits)
    ]
    opposite_mixed_direction_bound_stabs = [
        stab
        for stab in stabilizers
        if set(stab.data_qubits).issubset(opposite_direction_boundary_qubits)
    ]

    # Find the mixed direction vector
    mixed_direction_vector = direction_to_coord(diag_direction)

    anc_qubits_to_init = {"X": [], "Z": []}

    # Step 1
    # Find the first set of ancilla qubits to initialize
    for stab in stabilizers:
        if stab in mixed_direction_bound_stabs:
            # Skip same boundary stabilizers
            continue

        stab_anc = stab.ancilla_qubits[0]
        stab_type = stab.pauli_type

        other_anc = tuple(
            coord1 + coord2
            for coord1, coord2 in zip(stab_anc, mixed_direction_vector, strict=True)
        )

        anc_qubits_to_init[stab_type].append(other_anc)

    # Step 2
    # Find the teleportation qubits
    teleportation_qubit_pairs = []
    data_qubit_vector = direction_to_coord(
        diag_direction.opposite(), anc_qubit_sublattice_index
    )

    for stab in stabilizers:
        if stab in opposite_mixed_direction_bound_stabs:
            # Skip opposite boundary stabilizers
            continue
        stab_anc = stab.ancilla_qubits[0]
        pauli = stab.pauli_type

        # Check if it's a teleportation qubit
        if stab_anc not in anc_qubits_to_init[pauli]:
            anc_qubits_to_init[pauli].append(stab_anc)

            # Find its data qubit pair
            data_qubit = tuple(map(sum, zip(stab_anc, data_qubit_vector, strict=True)))
            teleportation_qubit_pairs.append((stab_anc, data_qubit))

    # Step 3
    # Find the data qubits to initialize
    data_qubits_to_init = {"X": [], "Z": []}
    # We need the ancilla to data qubit vector for that
    anc_to_data_vector = direction_to_coord(diag_direction, anc_qubit_sublattice_index)
    for stab in mixed_direction_bound_stabs:
        anc_qubit = stab.ancilla_qubits[0]
        diag_data_qubit = tuple(
            coord1 + coord2
            for coord1, coord2 in zip(anc_qubit, anc_to_data_vector, strict=True)
        )
        data_qubits_to_init[stab.pauli_type].append(diag_data_qubit)

    # Finally, relocate the qubits if needed
    if relocation_diag_direction is not None:
        anc_relocation_vector = direction_to_coord(
            relocation_diag_direction, anc_qubit_sublattice_index
        )
        data_relocation_vector = direction_to_coord(
            relocation_diag_direction, data_qubit_sublattice_index
        )

        def sum_vecs(vec1, vec2):
            return tuple(map(sum, zip(vec1, vec2, strict=True)))

        anc_qubits_to_init = {
            key: [sum_vecs(q, anc_relocation_vector) for q in val]
            for key, val in anc_qubits_to_init.items()
        }
        data_qubits_to_init = {
            key: [sum_vecs(q, data_relocation_vector) for q in val]
            for key, val in data_qubits_to_init.items()
        }
        teleportation_qubit_pairs = [
            tuple(
                (
                    sum_vecs(q, anc_relocation_vector)
                    if q[2] == 1
                    else sum_vecs(q, data_relocation_vector)
                )
                for q in q_pair
            )
            for q_pair in teleportation_qubit_pairs
        ]

    return anc_qubits_to_init, data_qubits_to_init, teleportation_qubit_pairs


def get_swap_qec_cnots(
    interpretation_step: InterpretationStep,
    new_block: RotatedSurfaceCode,
    move_diag_direction: DiagonalDirection,
    stab_schedule_dict: dict[Stabilizer, DetailedSchedule],
    anc_qubit_initialization: dict[str, list[tuple[int, ...]]],
    data_qubit_initialization: dict[str, list[tuple[int, ...]]],
) -> Circuit:
    """Generates the circuit for the final step of the swap-then-qec operation. This
    step moves the data qubits to their final positions where the new_block definition
    is.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step. Note that it may be mutated by generating new
        channels.
    new_block : RotatedSurfaceCode
        The new block after the diagonal move.
    move_diag_direction : DiagonalDirection
        The diagonal direction in which the block is moving.
    stab_schedule_dict : dict[Stabilizer, DetailedSchedule]
        A dictionary mapping each stabilizer to its detailed schedule.
    anc_qubit_initialization : dict[str, list[tuple[int, ...]]]
        A dictionary mapping each Pauli type to its list of ancilla qubit
        initializations.
    data_qubit_initialization : dict[str, list[tuple[int, ...]]]
        A dictionary mapping each Pauli type to its list of data qubit initializations.

    Returns
    -------
    Circuit
        The generated circuit for the final step of the swap-then-qec operation.
    """

    # Find the vectors to and from the final positions
    op_move_diag_direction = move_diag_direction.opposite()

    # Initialize the cnots and find the first layer of them from
    cnots = [[] for _ in range(4)]
    for q in new_block.data_qubits:
        q_prev = update_qubit_coords([q], op_move_diag_direction)[0]
        if q in anc_qubit_initialization["Z"] + data_qubit_initialization["Z"]:
            cnot_pair = (q_prev, q)
        elif q in anc_qubit_initialization["X"] + data_qubit_initialization["X"]:
            cnot_pair = (q, q_prev)
        else:
            raise ValueError(
                "Data qubit not found in either ancilla or data qubit initialization "
                "lists."
            )

        cnots[0].append(cnot_pair)

    # Find the rest of the cnots
    for stab in new_block.stabilizers:
        if len(stab.data_qubits) == 4:
            data_qubits = stab_schedule_dict[stab].get_stabilizer_qubits(stab)
        else:
            boundary_direction = next(
                dir for dir in Direction if stab in new_block.boundary_stabilizers(dir)
            )
            data_qubits = stab_schedule_dict[stab].get_stabilizer_qubits(
                stab, boundary_direction
            )
        for i, data_qubit in enumerate(data_qubits):
            if data_qubit is None or i == 0:
                # Skip if the data qubit is:
                # - None  (2-qubit stabilizer)
                # - the first qubit (already handled due to swap-qec)
                continue
            if stab.pauli_type == "Z":
                # If the stabilizer is Z, cnot from data qubit to ancilla
                cnot_pair = (data_qubit, stab.ancilla_qubits[0])
            elif stab.pauli_type == "X":
                # If the stabilizer is X, cnot from ancilla to data qubit
                cnot_pair = (stab.ancilla_qubits[0], data_qubit)
            else:
                raise ValueError("Unknown stabilizer type.")
            cnots[i].append(cnot_pair)

    # Generate the circuit containing the cnots
    swap_then_qec_cnots = Circuit(
        name=(
            f"Swap-then-QEC CNOTs for moving block {new_block.unique_label} towards "
            f"{move_diag_direction.value}"
        ),
        circuit=[
            [
                Circuit(
                    "cx",
                    channels=[
                        interpretation_step.get_channel_MUT(q) for q in qubit_pair
                    ],
                )
                for qubit_pair in cnot_slice
            ]
            for cnot_slice in cnots
        ],
    )

    return swap_then_qec_cnots


def generate_syndrome_measurement_circuit_and_cbits(
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    actual_anc_qubit_relocation_vector: tuple[int, ...] = (0, 0, 0),
) -> tuple[Circuit, list[tuple[Cbit, ...]]]:
    """
    Generate and return the circuit that measures the stabilizers of the block and
    the list of stabilizer measurements as tuples of Cbits.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step. Note that it may be mutated by generating new
        channels.
    block : RotatedSurfaceCode
        The block whose stabilizers are to be measured.
    actual_anc_qubit_relocation_vector : tuple[int, ...], optional
        The vector to the actual ancilla qubit, by default (0, 0, 0)

    Returns
    -------
    Circuit
        The circuit that performs the measurement of the stabilizers of the block.
        Note that it's just the final measurement operations and not the full
        stabilizer measurement circuit.
    list[tuple[Cbit, ...]]
        The list of stabilizer measurements as tuples of Cbits. The order of the list
        corresponds to the order of the stabilizers in block.stabilizers.
    """

    # Initialize the list of circuits
    stab_measurements = []
    meas_circ_seq = [[]]
    # Find the vector to the actual ancilla qubit
    for stab in block.stabilizers:
        # Find the data qubit containing the syndrome
        actual_anc_qubit = tuple(
            map(
                sum,
                zip(
                    stab.ancilla_qubits[0],
                    actual_anc_qubit_relocation_vector,
                    strict=True,
                ),
            )
        )

        actual_anc_channel = interpretation_step.get_channel_MUT(actual_anc_qubit)

        cbit = interpretation_step.get_new_cbit_MUT(f"c_{actual_anc_qubit}")
        stab_measurements.append((cbit,))

        cbit_channel = interpretation_step.get_channel_MUT(
            f"{cbit[0]}_{cbit[1]}", channel_type="classical"
        )

        m_circ = Circuit(
            f"measure_{stab.pauli_type}", channels=[actual_anc_channel, cbit_channel]
        )

        # Append the circuit to the list
        meas_circ_seq[0].append(m_circ)

    # Compile the stabilizer measurement circuit
    stab_meas_circ = Circuit(
        "measure stabilizer ancillas",
        circuit=meas_circ_seq,
    )

    return stab_meas_circ, stab_measurements


def generate_and_append_block_syndromes_and_detectors(
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    syndrome_measurement_cbits: list[tuple[Cbit, ...]],
) -> None:
    """Generate and append syndromes and detectors to the interpretation step.

    NOTE: This should probably be made into a method of InterpretationStep.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step. Note that it may be mutated by generating new
        channels.
    block : RotatedSurfaceCode
        The block to which the operation will be applied.
    syndrome_measurement_cbits : list[tuple[Cbit, ...]]
        The list of syndrome measurement classical bits. It has to correspond to the
        stabilizers of the block.
    """
    new_syndromes = generate_syndromes(
        interpretation_step=interpretation_step,
        stabilizers=block.stabilizers,
        block=block,
        stab_measurements=syndrome_measurement_cbits,
    )
    # Generate the new detectors for the new syndromes
    new_detectors = generate_detectors(interpretation_step, new_syndromes)
    # Append the syndromes and detectors to the interpretation step
    interpretation_step.append_syndromes_MUT(new_syndromes)
    interpretation_step.append_detectors_MUT(new_detectors)


def generate_teleportation_measurement_circuit_with_updates(
    interpretation_step: InterpretationStep,
    new_block: RotatedSurfaceCode,
    anc_qubits_to_init: dict[str, list[tuple[int, ...]]],
    teleportation_qubit_pairs: list[tuple[tuple[int, ...], tuple[int, ...]]],
) -> Circuit:
    """Generate the circuit that finalizes the teleportation operation. This includes
    the measurement of the data qubits and the updating of the necessary stabilizers
    based on the measurement outcomes.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step. Note that it may be mutated by generating new
        channels and updating stabilizers and logical operators.
    new_block : RotatedSurfaceCode
        The new block after the diagonal move.
    anc_qubits_to_init : dict[str, list[tuple[int, ...]]]
        A dictionary mapping each Pauli type to its list of ancilla qubit
        initializations.
    teleportation_qubit_pairs : list[tuple[tuple[int, ...], tuple[int, ...]]]
        A list of tuples, each containing a pair of qubits involved in the
        teleportation operation. The first qubit in each tuple is the ancilla qubit to
        be corrected, and the second qubit is the data qubit to be measured.

    Returns
    -------
    Circuit
        The circuit that finalizes the teleportation operation
    """
    teleportation_circ_seq = [[]]
    for corrected_qubit, measured_qubit in teleportation_qubit_pairs:
        # Obtain the necessary channels and cbit
        cbit = interpretation_step.get_new_cbit_MUT(f"c_{measured_qubit}")
        cbit_channel = interpretation_step.get_channel_MUT(
            f"{cbit[0]}_{cbit[1]}", channel_type="classical"
        )
        measured_qubit_channel = interpretation_step.get_channel_MUT(measured_qubit)

        # Determine the operations based on the initialization basis of the
        # ancilla qubit
        if corrected_qubit in anc_qubits_to_init["X"]:
            measure_op_name = "measure_z"
            update_pauli = "Z"
        elif corrected_qubit in anc_qubits_to_init["Z"]:
            measure_op_name = "measure_x"
            update_pauli = "X"
        else:
            raise ValueError("The ancilla qubit was not found in the initialization.")

        # Define the circuits that measure the data qubit and operate on the ancilla
        # qubit based on the measurement
        meas_circ = Circuit(
            measure_op_name,
            channels=[measured_qubit_channel, cbit_channel],
        )

        # Append the circuits to the list
        teleportation_circ_seq[0].append(meas_circ)

        # Do the same for the logical operator (it's going to be only one operator)
        logical_ops = (
            new_block.logical_z_operators
            if update_pauli == "Z"
            else new_block.logical_x_operators
        )
        logical_ops_involved = [
            op for op in logical_ops if corrected_qubit in op.data_qubits
        ]
        for op in logical_ops_involved:

            interpretation_step.update_logical_operator_updates_MUT(
                update_pauli, op.uuid, (cbit,), False
            )

        # Find all appropriate stabilizers
        stabs_to_update = [
            stab
            for stab in new_block.stabilizers
            if stab.pauli_type == update_pauli  # correct flavor
            and corrected_qubit in stab.data_qubits  # qubit part of stabilizer
        ]

        # Append the Cbit on the updates of the stabilizers
        for stab in stabs_to_update:
            current_upd = interpretation_step.stabilizer_updates.get(stab.uuid, ())
            interpretation_step.stabilizer_updates[stab.uuid] = current_upd + (cbit,)

    # Compile the teleportation circuit finalization
    teleportation_circuit_finalization = Circuit(
        f"teleportation measurements of block {new_block.unique_label}",
        circuit=teleportation_circ_seq,
    )

    return teleportation_circuit_finalization



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations
from functools import reduce
from typing import Optional, Union
from uuid import uuid4
import logging

from pydantic import Field, field_validator, ValidationInfo
from pydantic.dataclasses import dataclass

from .channel import Channel, ChannelType
from .utilities.serialization import apply_to_nested
from .utilities.validation_tools import (
    uuid_error,
    dataclass_config,
    distinct_error,
    ensure_tuple,
    retrieve_field,
    no_name_error,
)

logging.basicConfig(format="%(name)s - %(levelname)s - %(message)s")
log = logging.getLogger(__name__)


@dataclass(config=dataclass_config)
class Circuit:
    """
    A serializable, recursive circuit representation. Previously defined circuit
    structures can be either reused as nested circuit elements or identified by name
    only for compression

    Parameter
    ---------
    name: str
        The name of the circuit/operation, identifying its behaviour: e.g. "Hadamard",
        "CNOT" or "entangle" for the combination of a Hadamard and a CNOT gate.

    circuit: tuple[tuple[Circuit, ...], ...]
        Alternative inputs are: Circuit, list[Circuit], list[list[Circuit]], ...
        Pydantic type conversion accepts all inputs stated here:
        https://docs.pydantic.dev/latest/concepts/conversion_table/

        The list of nested circuit elements. The outer tuple represents the time step
        (tick) that the enclosed circuits are executed at. The inner tuple
        contains parallel circuits that are executed at the same time step. So each
        channel can only operated on by one circuit each tick.

        An input of a 1D list/tuple of circuits is interpreted as a sequence of
        circuits. So each circuit is executed at its own tick and after the execution
        of the previous circuit is complete.

    channels: tuple[Channel, ...]
        Alternative inputs are: Channel, list[Channel], set[Channel], ...
        Pydantic type conversion accepts all inputs stated in the link above.

        The list of channels involved in the circuit

    duration: Optional[int]
        Duration of the circuit in number of time steps. If not provided, defaults to
        ``None``.

    id: str
        The unique identifier of the circuit
    """

    name: str
    circuit: tuple[tuple[Circuit, ...], ...] = Field(
        default_factory=tuple, validate_default=True
    )
    channels: tuple[Channel, ...] = Field(default_factory=tuple, validate_default=True)
    duration: Optional[int] = Field(default=None, validate_default=True)
    description: str = Field(default="", validate_default=True)
    id: str = Field(default_factory=lambda: str(uuid4()))

    # Validation functions

    _validate_name = field_validator("name")(no_name_error)
    _validate_uuid = field_validator("id")(uuid_error)

    @field_validator("circuit", mode="before")
    @classmethod
    def format_circuit(
        cls, circuit: Union[Circuit, list[Circuit], list[list[Circuit]]]
    ):
        """
        Format the circuit field to be a 2D tuple of circuits.
        """
        if isinstance(circuit, Circuit):
            return ((circuit,),)
        if not circuit:  # base gate
            return ()
        # The default interpretation of a list of circuit input is sequential execution
        # of the gates in the list.
        if circuit and all(isinstance(gate, Circuit) for gate in circuit):
            # Add empty tuples to match the gates durations
            return reduce(
                lambda x, y: x + ((y,),) + ((),) * (y.duration - 1), circuit, ()
            )

        return circuit

    _validate_channels_tuple = field_validator("channels", mode="before")(ensure_tuple)

    @field_validator("circuit")
    @classmethod
    def validate_timing(
        cls, circuit: tuple[tuple[Circuit, ...], ...]
    ) -> tuple[tuple[Circuit, ...], ...]:
        """
        Validates, that all time steps and durations are consistent. I.e. that no two
        gates are scheduled on the same channel at the same time.
        """
        occupancy_dict = {}

        for tick, time_step in enumerate(circuit):
            for gate in time_step:
                for channel in gate.channels:
                    if occupancy_dict.get(channel.id, -1) < tick:
                        occupancy_dict[channel.id] = tick + gate.duration - 1
                    else:
                        raise ValueError(
                            "Error while setting up composite circuit: Channel "
                            f"{channel.label}({channel.id[0:6]}..) is subject to more "
                            f"than one operation at tick {tick}."
                        )

        return circuit

    @field_validator("channels")
    @classmethod
    def adjust_channels(cls, channels: Union[Channel, list[Channel]], values: dict):
        """
        Adjusts the channels of the circuit based on the channels of the
        nested circuits.

        Parameters
        ----------
        channels : Union(list[Channel], Channel)
            The channels of the circuit.

        values : dict
            Values of other fields of the Circuit object.

        Returns
        -------
        list[Channel]
            The adjusted list of channels of the circuit.
        """

        def derive_channels(circuit: tuple[tuple[Circuit, ...], ...]):
            """
            Derive channels from nested circuits when the `channels` field is empty.
            The order of the channels of each type is not conserved (or important).
            The output channels are ordered by type.
            """
            all_channels = {
                channel
                for tick in circuit
                if tick
                for circ in tick
                for channel in circ.channels
            }
            # Order channels by type
            typing_order = (
                ChannelType.QUANTUM,
                ChannelType.CLASSICAL,
            )
            ordered_channels = sorted(
                all_channels, key=lambda x: typing_order.index(x.type)
            )
            return tuple(ordered_channels)

        circuit = retrieve_field("circuit", values)

        match (len(circuit) == 0, len(channels) == 0):
            # (True, _) if base gate (no nested circuits)
            case (True, True):
                return ()
            case (True, False):
                return distinct_error(channels)
            case (False, True):
                return derive_channels(circuit)
            case (False, False):
                if set(derive_channels(circuit)) != set(channels):
                    raise ValueError(
                        "\nError while setting up composite circuit: Provided channels"
                        " do not match the channels of the sub-circuits. \nMake sure"
                        " that the sub-circuits channel ids and types match the ones"
                        " provided.\n"
                    )

        return channels

    @field_validator("duration")
    @classmethod
    def adjust_duration(cls, duration: int, info: ValidationInfo) -> int:
        """
        Sets the duration of the circuit based on the duration of the nested circuits
        (if any).
        """

        def derive_duration(circuit: tuple[tuple[Circuit, ...], ...]) -> int:
            """
            Derive the duration of the circuit from the nested circuits: calculate the
            end tick of each operation and return the maximum.
            """
            return max(
                # flatten to a 1D list
                reduce(
                    lambda x, y: x + y,
                    map(
                        # calculate the end tick of each operation
                        lambda tick: [op.duration + tick[0] for op in tick[1]],
                        enumerate(circuit),
                    ),
                    [],
                )
                + [len(circuit)]  # Include empty time steps length if it's larger
            )

        circuit = retrieve_field("circuit", info)

        match (not circuit, duration is None):
            # (True, _) if base gate (no nested circuits)
            case (True, True):
                return 1
            case (True, False):
                if duration < 1:
                    raise ValueError("Duration must be a positive integer.")
            case (False, True):
                return derive_duration(circuit)
            case (False, False):
                derived_duration = derive_duration(circuit)
                if derived_duration != duration:
                    raise ValueError(
                        f"Error while setting up composite circuit: Provided duration "
                        f"({duration}) does not match the duration of the sub-circuits "
                        f"({derived_duration})."
                    )

        return duration

    # Methods

    @classmethod
    def as_gate(
        cls, name: str, nr_qchannels: int, nr_cchannels: int = 0, duration: int = 1
    ):
        """
        Create a base gate by specifying the name and optionally the number of quantum
        and classical channels and the duration.

        Parameters
        ----------
        name : str
            The name of the gate.

        nr_qchannels : int
            The number of quantum channels it acts on.

        nr_cchannels : int
            The number of classical channels it acts on.

        duration : int
            The duration of the base gate.

        Returns
        -------
        Circuit
            The base gate Circuit object.
        """
        qchannels = [Channel(type=ChannelType.QUANTUM) for _ in range(nr_qchannels)]
        cchannels = [Channel(type=ChannelType.CLASSICAL) for _ in range(nr_cchannels)]
        return cls(name, channels=qchannels + cchannels, duration=duration)

    @classmethod
    def from_circuits(cls, name: str, circuit=None):
        """
        Create a Circuit object from a list of circuits with relative qubit indices.

        Parameters
        ----------
        name : str
            The name of the circuit.

        circuit : list[tuple[Circuit, list[int]]], list[list[tuple[Circuit, list[int]]]]
            The list of circuits with relative qubit indices.

        Returns
        -------
        Circuit
            The Circuit object.
        """

        def make_chan(
            cidx: int, ctype: ChannelType, cmap: dict[str, Channel]
        ) -> Channel:
            """
            Create a Channel for a relative index if it does not exist in the mapping
            dictionary yet. Check for Channel type inconsistencies of relative indices
            used in multiple sub-Circuits.
            """
            if cidx in cmap.keys():
                if cmap[cidx].type != ctype:
                    raise ValueError(
                        f"Provided channel indices are not consistent with respect"
                        f" to their types. Offending channel {cidx} has type {ctype}"
                        f" but has previously been used with a channel of type "
                        f"{cmap[cidx].type}."
                    )
                return cmap[cidx]
            new_chan = Channel(type=ctype)
            cmap[cidx] = new_chan
            return new_chan

        def make_circ(circtup: tuple[Circuit, list[int]], cmap: dict) -> Circuit:
            """
            Build a sub-Circuit with the correct Channels based on the given relative
            indices.
            """
            nr_prov_channels = len(circtup[1])
            nr_circ_channels = len(circtup[0].channels)

            if nr_prov_channels != nr_circ_channels:
                raise ValueError(
                    f"Provided number of channels {nr_prov_channels} does not match the"
                    f" number of channels {nr_circ_channels} in circuit "
                    f"{circtup[0].name}."
                )

            new_channels = [
                make_chan(cidx, ctype, cmap)
                for cidx, ctype in zip(
                    circtup[1],
                    map(lambda chan: chan.type, circtup[0].channels),
                    strict=True,
                )
            ]

            return circtup[0].clone(new_channels)

        if circuit is None:
            circuit = []
        if isinstance(circuit, tuple) or len(circuit) < 2:
            raise ValueError(
                "Error while creating circuit via from_circuit(): The circuit must be "
                "a list of circuits. If the intention is to copy a circuit to deal "
                "with Channel objects directly, use the clone() method instead."
            )

        cmap = {}
        new_circ = apply_to_nested(circuit, make_circ, cmap)
        return cls(name, new_circ)

    def clone(self, channels: list[Channel] | None = None) -> Circuit:
        """
        Convenience method to clone a circuit structure that was defined before.

        Parameter
        ---------
        channels: list[Channel]
            Channels of the new circuit.

        Returns
        -------
        Circuit
        """

        def make_channel_map(
            old_channels: list[Channel], new_channels: list[Channel]
        ) -> dict[str, Channel]:
            """
            Create a mapping from old channel ids to new channels.

            Parameter
            ---------
            old_channels: list[Channel]
                The old channels to be mapped.

            new_channels: list[Channel]
                The new channels to be mapped.

            Returns
            -------
            dict[str, Channel]
                A dictionary that maps the old channel ids to the new channels.
            """

            def match_channel(
                old_channel: Channel, index: int, new_channels: list[Channel]
            ) -> Channel:
                if index < len(new_channels):
                    if old_channel.is_quantum() != new_channels[index].is_quantum():
                        raise ValueError(
                            "Error while cloning circuit: CLASSICAL channels cannot be"
                            " assigned to QUANTUM channels and vice versa."
                        )
                    return new_channels[index]
                return Channel(type=old_channel.type)

            new_channels = list(
                map(
                    lambda old_channel: match_channel(
                        old_channel[1], old_channel[0], new_channels
                    ),
                    enumerate(old_channels),
                )
            )
            old_ids = list(map(lambda channel: channel.id, old_channels))

            return dict(zip(old_ids, new_channels, strict=True))

        def update_sub_circuit(
            circuit: Circuit, channel_map: dict[str, Circuit]
        ) -> Circuit:
            """
            Update the nested sub-circuits with new channels recursively.

            Parameter
            ---------
            circuit: Circuit
                The circuit to be updated.

            channel_map: dict[str, Channel]
                The channel map used for looking up the new circuit's channels.

            Returns
            -------
            Circuit
                The updated circuit.
            """
            new_channels = [channel_map[channel.id] for channel in circuit.channels]

            # If this is an IfElseCircuit (detected via the marker), rebuild it
            # by recursively updating its branches. We avoid importing the
            # IfElseCircuit class here to prevent circular imports by using
            # `type(circuit)` to construct the new instance.
            if hasattr(circuit, "_loom_ifelse_marker"):
                # update branches recursively
                new_if = update_sub_circuit(circuit.if_circuit, channel_map)
                new_else = update_sub_circuit(circuit.else_circuit, channel_map)
                new_cond = update_sub_circuit(circuit.condition_circuit, channel_map)

                # construct new IfElseCircuit using the actual runtime class
                return type(circuit)(
                    if_circuit=new_if, else_circuit=new_else, condition_circuit=new_cond
                )

            # Regular Circuit: rebuild with updated subcircuits and channels
            return Circuit(
                circuit.name,
                tuple(
                    tuple(update_sub_circuit(circ, channel_map) for circ in tick)
                    for tick in circuit.circuit
                ),
                new_channels,
                circuit.duration,
                # channels can't be inferred here, since order of channels in this list
                # matters.
            )

        if channels is None:
            channels = []
        if isinstance(channels, Channel):
            channels = [channels]
        channel_map = make_channel_map(self.channels, channels)
        return update_sub_circuit(self, channel_map)

    def nr_of_qubits_in_circuit(self):
        """
        Returns the number of qubits used in the circuit across all branches.

        Parameters
        ----------
        circuit : Circuit
            recursive graph circuit representation

        Returns
        -------
        int
            the number of qubits in the circuit
        """
        return len(list(filter(lambda channel: channel.is_quantum(), self.channels)))

    def circuit_seq(self):
        """
        Returns the sequence of sub-circuits in the circuit field.

        Returns
        -------
        tuple[Circuit, ...]
            The list of sub-circuits in sequence, disregarding ticks.
        """
        return reduce(lambda x, y: x + y, self.circuit, ())

    def flatten(self) -> Circuit:
        """
        Returns the flattened circuit as a copy where all elements in the circuit
        list are physical operations, and there is no further nesting.

        Returns
        -------
        Circuit
            The flattened circuit
        """
        flat_circuit = []

        # This is the Depth First Search (DFS) traversal of a tree:
        # https://en.wikipedia.org/wiki/Tree_traversal#Depth-first_search
        queue = [self]
        while len(queue) > 0:
            next_circuit = queue.pop()
            if len(next_circuit.circuit) == 0:  # If the circuit does not contain
                # subcircuits, it must be a physical gate and it is added to the
                # flattened circuit array
                flat_circuit.append(next_circuit)
            elif hasattr(next_circuit, "_loom_ifelse_marker"):  # IfElseCircuit marker
                flat_circuit.append(next_circuit.flatten())
            else:
                for tick in next_circuit.circuit:
                    for circ in tick:
                        if circ != ():  # If it is not an empty tuple
                            queue.append(circ)  # Add subcircuits to queue
        # The circuit list needs to be reversed because of the last in first out queue
        flat_circuit.reverse()

        return Circuit(
            self.name,
            circuit=flat_circuit,
            channels=self.channels,
        )

    @classmethod
    def unroll(cls, circuit: Circuit) -> tuple[tuple[Circuit, ...], ...]:
        """
        Unrolls the circuits within the time slices using a Depth First Search
        algorithm until the final sequence is composed of only base gates. This method
        preserves the time structure of the circuit (unlike flatten).
        Note that this method returns the unrolled circuit sequence, not a new Circuit.

        Returns
        -------
        tuple[tuple[Circuit, ...], ...]
            The unrolled circuit time sequence
        """
        unrolled_circuit_time_sequence = [
            () for _ in range(max(len(circuit.circuit), circuit.duration))
        ]
        # Contains the current sub-circuits that need to be unrolled and the time index
        # within unrolled_circuit_time_sequence
        stack = [(0, circuit)]

        # Traverse the recursive circuit using a Depth First Search algorithm
        while stack:
            time, circ = stack.pop()
            # If the circuit is empty, it is a base gate and can be
            # added to the final sequence
            if not circ.circuit:  # Base gate
                unrolled_circuit_time_sequence[time] += (circ,)
            # And if the circuit is an IfElseCircuit, use child unroll method
            elif hasattr(circ, "_loom_ifelse_marker"):
                unrolled_circuit_time_sequence[time] += circ.unroll(circ)
            # Else it's a composite circuit and is added to the stack
            # with the associated time index
            else:
                for i, tick in enumerate(circ.circuit):
                    for sub_circ in reversed(tick):
                        stack.append((time + i, sub_circ))
        return tuple(unrolled_circuit_time_sequence)

    # pylint: disable=too-many-return-statements
    def __eq__(self, other) -> bool:
        """
        Check whether two circuits perform the same gate sequence. I.e. check if the
        same gates are applied to the same qubits in the same order. Circuit and qubit
        names are ignored. It only matters that gates are applied to the same qubits, no
        matter what their internal id or their label is. Any nested structure of the
        circuits is ignored, i.e. the two circuits are unrolled before comparison.
        Note that the order of gates within a timeslice does not matter. It can be
        checked for if one compares a tuple of tuples of gates (Circuit.circuit)
        instead of using `Circuit.__eq__`. The order of the timeslices themselves does
        matter. Empty timeslices are also taken into account.

        Note that this overwrites the default `__eq__` method which would check for
        exact equality, including the equality of all uuids. There are only very few
        cases where one would need to check for exact equality including equality of
        uuids. If such a function will every be needed, it should be implemented as a
        separate method like `is_identical(self, other)` or similar. Since checking for
        equality but ignoring the uuids is the much common use case, overwriting the
        == operator for this check is the better default.

        Returns
        -------
        bool
            True if the two circuits are equivalent, False otherwise
        """
        # Create a mapping from channel ids from the first circuit to the channel ids
        # of the second circuit. This dict is constructed iteratively in the for loop.
        # For every gate, if the included channels are not contained in the map yet,
        # they are added to the map. If the channel is already in the map, the channel
        # of the second circuit is checked to correspond to the same qubit as the
        # channel of the first circuit. If not, the circuits are not equivalent.

        # We allow uneven lengths in the zips because this is an equality check
        channel_map = {}
        if isinstance(other, Circuit):
            # Unroll the circuits to a tuple of tuples of base gates
            circ_sequence1 = Circuit.unroll(self)
            circ_sequence2 = Circuit.unroll(other)
            if len(circ_sequence1) != len(circ_sequence2):
                log.info("The two circuits have a different number of time slices.")
                log.debug("%s != %s\n", len(circ_sequence1), len(circ_sequence2))
                return False

            # Check every time slice of the two circuits
            for time_step, (time_slice1, time_slice2) in enumerate(
                zip(circ_sequence1, circ_sequence2, strict=False)
            ):
                if len(time_slice1) == 0 and len(time_slice2) == 0:
                    continue  # Both time slices are empty tuples

                if len(time_slice1) != len(time_slice2):
                    log.info(
                        "The two circuits have a different number of gates in a "
                        "time slice."
                    )
                    log.debug(
                        "%s != %s for time slices %s and %s\n",
                        len(time_slice1),
                        len(time_slice2),
                        time_slice1,
                        time_slice2,
                    )
                    return False  # Unequal tuple length

                # Sort the gates by name within a time slice to compare them
                for gate1, gate2 in zip(
                    sorted((gate for gate in time_slice1), key=lambda x: x.name),
                    sorted((gate for gate in time_slice2), key=lambda x: x.name),
                    strict=False,
                ):
                    if hasattr(gate1, "_loom_ifelse_marker") or hasattr(
                        gate2, "_loom_ifelse_marker"
                    ):
                        return gate1 == gate2
                    # The two timeslices must have the same gates (names)
                    if gate1.name != gate2.name:
                        log.info(
                            "The two circuits have different gates in a time slice."
                        )
                        log.debug(
                            "For time steps %s: %s and %s: %s, \n"
                            "    %s != %s for gates %s and %s\n",
                            time_step,
                            time_slice1,
                            time_step,
                            time_slice2,
                            gate1.name,
                            gate2.name,
                            gate1,
                            gate2,
                        )
                        return False

                    # Check whether the channels are the same.
                    # This is done by comparing the sets of channel ids of the two
                    # circuits where for the first circuit, the ids are translated to
                    # the ids of the second circuit using the channel map
                    for ch1, ch2 in zip(gate1.channels, gate2.channels, strict=False):
                        if ch1.id not in channel_map:
                            channel_map[ch1.id] = ch2.id
                    if [ch.id for ch in gate2.channels] != [
                        channel_map.get(ch.id) for ch in gate1.channels
                    ]:
                        log.info("The two circuits have different channels in a gate.")
                        log.debug(
                            "\n    %s\n        !=\n    %s\n",
                            [(ch.type, ch.label) for ch in gate2.channels],
                            [(ch.type, ch.label) for ch in gate1.channels],
                        )
                        return False

            # No differences found, the circuits are equivalent
            return True

        # Else, cannot compare the circuit for equivalence with another object with is
        # not a `Circuit`
        return NotImplemented

    def __repr__(self):
        n_ticks = len(self.circuit)
        # Construct the title string
        if n_ticks == 0:
            # If the circuit is empty, it is a base gate
            title = f"{self.name} (base gate)\n"
        else:
            # If the circuit is not empty, it is a composite circuit
            # The title is the name of the circuit and the number of ticks
            title = f"{self.name} ({n_ticks} ticks)\n"
        tick_str = title
        for i, tick in enumerate(self.circuit):
            # Omit ticks occupied by empty tuples or lower level circuits
            # (i.e. not base gates)
            if len(tick) != 0:
                for gate in tick:
                    if hasattr(gate, "_loom_ifelse_marker"):
                        tick_str += f"{i}: {gate.__repr__()}\n"
                    else:
                        tick_str += f"{i}: " + gate.name + "\n"
        # Delete the last newline character
        tick_str = tick_str[:-1]
        return tick_str

    def detailed_str(self):
        """
        Detailed string representation for a `Circuit`, displaying the gates
        and channels per tick.
        """
        tick_str = f"{self.name}\n"
        for i, tick in enumerate(self.circuit):
            tick_str += f"{i}: "
            for gate in tick:
                if hasattr(gate, "_loom_ifelse_marker"):
                    tick_str += gate.detailed_str()
                else:
                    tick_str += f"{gate.name} - "
                    tick_str += f"{' '.join(str(chan.label) for chan in gate.channels)}"
                    tick_str += "\n"
        return tick_str

    @staticmethod
    def construct_padded_circuit_time_sequence(
        circuit_time_sequence: tuple[tuple[Circuit, ...], ...],
    ) -> tuple[tuple[Circuit, ...], ...]:
        """
        Construct a padded circuit time sequence.

        The input is a tuple of tuples of circuits, where each tuple of circuits
        represents a time step. Each time step may be of variable duration.
        The output is a tuple of tuples of circuits that includes empty tuples which
        represent time steps where the circuit is busy because of a composite
        sub-circuit.

        Note that the scheduling is done following the time structure of the input,
        if two composite circuits exist in the same time step, they will start at the
        same time but may end at different times. If there are conflicts between
        subsequent circuits, add the minimum amount of padding such that the circuit
        can be executed. The last composite circuit's padding will automatically be
        added since it is the last element in the sequence.

        E.g.:

        .. code-block:: python

            hadamard = Circuit("hadamard", channels=channels[0], duration=1)
            cnot = Circuit("cnot", channels=channels[0:2], duration=2)
            circuit_time_sequence = (
                (hadamard),
                (cnot,),
            )

        Constructing the padded circuit time sequence would result in:

        .. code-block:: python

            padded_circuit_time_sequence = (
                (hadamard,),
                (cnot,),
                (),
            )

        Similarly, if the input is:

        .. code-block:: python

            circuit_time_sequence = (
                (cnot),
                (hadamard,),
            )

        The padded circuit time sequence would be:

        .. code-block:: python

            padded_circuit_time_sequence = (
                (cnot,),
                (),
                (hadamard,),
            )

        To illustrate two circuits that are executed at the same time, but of variable
        duration:

        .. code-block:: python

            hadamard_2 = Circuit("hadamard_2", channels=channels[2], duration=1)
            circuit_time_sequence = (
                (cnot, hadamard_2,)
            )

        The padded circuit time sequence would be:

        .. code-block:: python

            padded_circuit_time_sequence = (
                (cnot, hadamard_2,),
                (),
            )

        where the cnot would span two time steps and hadamard_2 only one.

        Parameters
        ----------
        circuit_time_sequence : tuple[tuple[Circuit, ...], ...]
            The circuit time sequence to be padded.

        Returns
        -------
        tuple[tuple[Circuit, ...], ...]
            The padded circuit time sequence.
        """
        # Create a new time sequence
        padded_circuit_time_sequence = ()

        # Keep track of occupied channels and for how long they are being occupied
        occupancy_dictionary = {}
        for tick in circuit_time_sequence:
            # Find the occupancy of the current tick
            current_tick_occupancy = {
                channel.label: circuit.duration
                for circuit in tick
                for channel in circuit.channels
            }
            # Find the channels that are occupied both in the current tick and the
            # previous ticks
            conflicting_channels = set(occupancy_dictionary.keys()).intersection(
                set(current_tick_occupancy.keys())
            )
            # If there are conflicting channels, add padding accounting for the minimum
            # time required to remove conflicts and define the duration to deduct
            # from the occupancy_dictionary
            if conflicting_channels:
                duration = max(
                    occupancy_dictionary[channel] for channel in conflicting_channels
                )
                padded_circuit_time_sequence += ((),) * (duration - 1)
            # If there are no conflicting channels, the duration is 1
            else:
                duration = 1

            # Add the current tick after the padding
            padded_circuit_time_sequence += (tick,)

            # Free channels in the current tick: i.e. channels that are still in use
            # with gates from previous ticks, but are not involved with the current tick
            # (no conflict). Their duration is being counted down in the occupancy
            # dictionary.
            free_channels = set(occupancy_dictionary.keys()).difference(
                set(current_tick_occupancy.keys())
            )
            # Update the occupancy dictionary:
            # We remove the free channels that belong to completed gates and add the new
            # duration of the ones that are still occupied
            for channel in free_channels:
                if (new_duration := occupancy_dictionary.pop(channel) - duration) > 0:
                    occupancy_dictionary[channel] = new_duration
            # Add the newly occupied channels
            occupancy_dictionary.update(current_tick_occupancy)

        # Add the padding for the last tick
        if occupancy_dictionary:
            duration = max(occupancy_dictionary.values())
            padded_circuit_time_sequence += ((),) * (duration - 1)

        return padded_circuit_time_sequence



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import numpy as np


def g(x1: int, z1: int, x2: int, z2: int) -> int:
    """
    The g function as described in Aaronson's paper but written with
    bitwise operations. The result is the exponent of the imaginary
    unit accompanying the multiplication result of the two paulis P_2 * P_1.
    Example:
    X * Y = X * i Z X = -i Z
    So input (1,1,1,0) gives -1.

    Reference: https://arxiv.org/abs/quant-ph/0406196
    """
    return (
        (x1 | z1)
        * (1 - 2 * (x1 & z1))
        * (x1 * z2 * (2 * x2 - 1) + z1 * x2 * (1 - 2 * z2))
    )  # makes product zero for x1=z1=0  # flips sign for x1=z1=1


def g_npfunc(
    x1: np.ndarray, z1: np.ndarray, x2: np.ndarray, z2: np.ndarray
) -> np.ndarray:
    """
    Vectorized g function.

    Parameters
    ----------
    x1 : np.ndarray
        The x bits of the first pauli string.
    z1 : np.ndarray
        The z bits of the first pauli string.
    x2 : np.ndarray
        The x bits of the second pauli string.
    z2 : np.ndarray
        The z bits of the second pauli string.

    Returns
    -------
    np.ndarray
        The g values.
    """
    return np.frompyfunc(g, 4, 1)(x1, z1, x2, z2)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations
from uuid import uuid4, UUID

import numpy as np
from loom.eka.utilities import (
    paulichar_to_xz_npfunc,
    paulixz_to_char_npfunc,
    sparse_formatter,
)


class PauliFrame:
    """
    A class representing a Pauli Frame.
    """

    def __init__(
        self,
        x: np.ndarray,
        z: np.ndarray,
        direction: str = "forward",
        id: str = None,  # pylint: disable=redefined-builtin
    ) -> None:
        """Initializes the the Pauli Frame."""
        z = np.array(z)
        x = np.array(x)

        if len(x) != len(z):
            raise ValueError("The length of z and x vectors should be equal.")

        if not np.all((z == 0) | (z == 1)):
            raise ValueError("z vector should contain only 0 and 1.")

        if not np.all((x == 0) | (x == 1)):
            raise ValueError("x vector should contain only 0 and 1.")

        # broadcast to int8
        self.z = np.array(z, dtype=np.int8)
        self.x = np.array(x, dtype=np.int8)

        # Check validity of direction
        if direction not in ["forward", "backward"]:
            raise ValueError(
                f"Invalid direction '{direction}'. Must be 'forward' or 'backward'."
            )
        self.direction = direction
        # Check the validity of the ID
        self.id = str(uuid4()) if id is None else id
        try:
            uuid_obj = UUID(self.id)
        except ValueError as exc:
            raise ValueError(
                f"Invalid uuid: {self.id}. UUID must be version 4."
            ) from exc
        if uuid_obj.version != 4:
            raise ValueError(f"Invalid uuid: {self.id}. UUID must be version 4.")

    def __repr__(self) -> str:
        str_array = paulixz_to_char_npfunc(self.x, self.z)
        paulistr = "".join(str_array)
        return f"PauliFrame: {paulistr}"

    def __len__(self):
        return len(self.x)

    def copy(self) -> PauliFrame:
        """
        Returns a copy of the Pauli Frame.
        """
        return PauliFrame(self.x.copy(), self.z.copy())

    def __eq__(self, o_pf: PauliFrame) -> bool:
        # check if they have the same length
        if len(self) != len(o_pf):
            return False
        # return whether x's and z's match
        return np.all(self.x == o_pf.x) and np.all(self.z == o_pf.z)

    @classmethod
    def from_string(
        cls,
        pauli_string: str,
        direction: str = "forward",
        id: str = None,  # pylint: disable=redefined-builtin
    ) -> PauliFrame:
        """
        Create a PauliFrame from a string representation of a Pauli operator.
        """
        ps_array = np.array(list(pauli_string))

        x, z = paulichar_to_xz_npfunc(ps_array)

        return PauliFrame(x, z, direction, id)

    def sparse_format(self) -> dict:
        """
        Returns the sparse format of the Pauli Frame.
        """
        s = repr(self)
        return sparse_formatter({"+" + s[12:]})



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations
from uuid import uuid4

from pydantic.fields import Field
from pydantic.dataclasses import dataclass
from loom.eka.utilities import dataclass_config

from .utilities import Cbit


@dataclass(config=dataclass_config)
class Syndrome:
    """
    A syndrome is the measurement result of a stabilizer. This dataclass does not
    contain the actual value of the syndrome but it stores how to calculate the syndrome
    value based on the output one gets from executing the quantum circuit.

    Parameters
    ----------
    stabilizer : str
        Uuid of the stabilizer that this syndrome corresponds to.
    measurements : tuple[Cbit, ...]
        List of classical bits from which the syndrome is calculated. The syndrome is
        calculated by taking the parity of these measurements. This includes the readout
        of ancilla qubits as well as potential updates/corrections based on data qubit
        readouts during e.g. split or shrink operations.
    block : str
        UUID of the block which contains the stabilizer that is measured in this
        syndrome.
    round : int
        Syndrome extraction round of the respective block in which this syndrome was
        measured. Note that the rounds of syndrome extraction are counted for each block
        individually.
    corrections: tuple[Cbit, ...]
        List of classical bits that needed to account for parity computation between
        subsequent rounds of syndrome extraction. This is necessary when the stabilizer
        was modified during a lattice surgery operation
    uuid : str
        Unique identifier of the syndrome, created with the uuid module.
    """

    stabilizer: str
    measurements: tuple[Cbit, ...]
    block: str
    round: int
    corrections: tuple[Cbit, ...] = Field(default_factory=tuple)
    labels: dict[str, str | tuple[int, ...] | int] = Field(default_factory=dict)
    uuid: str = Field(default_factory=lambda: str(uuid4()), validate_default=True)

    def __eq__(self, other: Syndrome) -> bool:
        if not isinstance(other, Syndrome):
            raise NotImplementedError(
                f"Cannot compare Syndrome with {type(other)} object."
            )
        return (
            self.stabilizer == other.stabilizer
            and set(self.measurements) == set(other.measurements)
            and set(self.corrections) == set(other.corrections)
            and self.block == other.block
            and self.round == other.round
        )

    def __repr__(self) -> str:
        return (
            f"Syndrome(Measurements: {self.measurements}, "
            f"Corrections: {self.corrections}, Round: {self.round}, "
            f"Labels: {self.labels})"
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.stabilizer,
                self.measurements,
                self.block,
                self.round,
                self.corrections,
            )
        )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Stabilizer, Block

from ..syndrome import Syndrome
from ..interpretation_step import InterpretationStep
from ..utilities import Cbit


def generate_syndromes(
    interpretation_step: InterpretationStep,
    stabilizers: tuple[Stabilizer, ...],
    block: Block,
    stab_measurements: tuple[tuple[Cbit, ...], ...],
) -> tuple[Syndrome, ...]:
    """
    Generate new Syndromes for the given stabilizers and its associated block.
    Stabilizers are passed explicitly as they follow the same order as the
    `stab_measurements` variable used to compute the Syndromes.

    CAUTION: This function pops the entries from the stabilizer_updates field of the
    interpretation step to compute corrections. This may cause issues in the future if
    the information in this field also needs to be accessed somewhere else.


    Parameters
    ----------
    interpretation_step : InterpretationStep
        The updated interpretation step implementing the operation
    stabilizers : tuple[Stabilizer, ...]
        Stabilizers that were measured, results are included in `stab_measurements`
    block : Block
        Block containing the stabilizers
    stab_measurements : tuple[tuple[Cbit, ...], ...]
        Measurements used to create the syndromes. Each index contains a tuple of Cbits
        associated to the stabilizer at the same index in `stabilizers`.

    Returns
    -------
    tuple[Syndrome, ...]
        Syndromes created for the stabilizers, they are returned in the same order as
        the stabilizers are given.
    """
    # Find the round that needs to be associated with the new syndromes
    # If the block exists in block_qec_rounds, then we increment the round
    if block.uuid in interpretation_step.block_qec_rounds.keys():
        new_round = interpretation_step.block_qec_rounds[block.uuid]  # Get the index
        interpretation_step.block_qec_rounds[block.uuid] += 1  # Then increment
    # If the block does not exist in block_qec_rounds, then we create it
    else:
        new_round = 0
        interpretation_step.block_qec_rounds[block.uuid] = 1

    # Extract stabilizer labels to be inherited by the Syndromes and add the time stamp
    time_stamp = interpretation_step.get_timestamp()

    # Create new Syndromes
    new_syndromes = tuple(
        Syndrome(
            stabilizer=stabilizer.uuid,
            measurements=measurements,
            block=block.uuid,
            round=new_round,
            labels=block.get_stabilizer_label(stabilizer.uuid)
            | {"time_coordinate": (time_stamp,)},
            corrections=interpretation_step.stabilizer_updates.pop(
                stabilizer.uuid, tuple()
            ),  # get and remove the correction from int_step.stabilizer_updates
        )
        for stabilizer, measurements in zip(stabilizers, stab_measurements, strict=True)
    )

    return new_syndromes



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from .base_applicator import BaseApplicator
from .code_applicator import CodeApplicator
from .conditional_logical_pauli import conditional_logical_pauli
from .measure_block_syndromes import measureblocksyndromes
from .measure_logical_pauli import measurelogicalpauli
from .logical_pauli import logical_pauli
from .reset_all_data_qubits import reset_all_data_qubits
from .reset_all_ancilla_qubits import reset_all_ancilla_qubits
from .generate_detectors import generate_detectors
from .generate_syndromes import generate_syndromes
